import os
import shutil
from enum import Enum
from typing import Any

ABS_PATH = os.path.dirname(os.path.abspath(__file__))


class Theorem(Enum):
    """
    Theorem is a class which is used to define the type of theorems that can be used in the solver.
    """
    FARKAS = 'farkas'
    HANDELMAN = 'handelman'
    PUTINAR = 'putinar'


class AvailabilityDict(dict):
    """
    AvailabilityDict is a class which is used to define a dictionary that can
    be used to hierarchically check the availability of commands in the system,
    given an array of possible commands.
    
    Parameters
    ----------
    *args : Any
        The arguments that are passed to the dictionary.
    **kwargs : Any
        The keyword arguments that are passed to the dictionary.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for key, value in self.items():
            if not isinstance(value, list):
                self[key] = [value]

    def __getitem__(self, key: Any) -> Any:
        """
        Retrieve the first available command from the list of possible commands.

        Parameters
        ----------
        key : Any
            The key to be retrieved from the dictionary.

        Returns
        -------
        Any
            The first available command from the list of possible commands.
        """
        possible_values = super().__getitem__(key)
        for value in possible_values:
            if AvailabilityDict.available(value):
                return value
        return None

    def __setitem__(self, key: Any, value: Any) -> None:
        if not isinstance(value, list):
            value = [value]
        super().__setitem__(key, value)

    @staticmethod
    def available(item):
        return shutil.which(item) is not None


class Constant:
    """
    This class consist of some constant dictionaries which are used for configuration of the solvers.
    """
    options = {
        'z3': '(set-option :print-success false)\n' +
              '(set-option :produce-models true)\n',
        'mathsat': '(set-option :print-success false)\n' +
                   '(set-option :produce-models true)\n',
        'default': ''
    }

    default_path = AvailabilityDict({
        'z3': ['z3', os.path.join(ABS_PATH, '..', '..', 'solver', 'z3')],
        'mathsat': ['mathsat', os.path.join(ABS_PATH, '..', '..', 'solver', 'mathsat')],
        'default': ''
    })

    command = {
        'z3': '',
        'mathsat': '',
        'default': ''
    }



from __future__ import annotations

class UnknownVariable:
    """
    This class represents a variable that is not known yet.
    
    Parameters
    ----------
    name : str, optional
        The name of the variable. By default None.
    type_of_var : str, optional
        The type of the variable. By default None.
        
    Attributes
    ----------
    id : int
        The unique number assigned to each new generated variable.
    name : str
        The given name to each variable.
    type : str
        The type of the variable. It can be referred as the reason why the variable is generated.
        For example, is it a template variable or it is generated by farkas algorithm.
    number_of_variables : int
        The number of generated variables to find new id for the recently generated variable
    """
    number_of_variables = 0

    @staticmethod
    def get_new_id() -> int:
        """ 
        Get new unique id for the newest variable

        Returns
        -------
        int
            A number to assign to the latest created variable
        """
        UnknownVariable.number_of_variables += 1
        return UnknownVariable.number_of_variables

    def __init__(self, name: str = None, type_of_var: str = None):
        self.name = name
        self.type = type_of_var
        self.id = UnknownVariable.get_new_id()

    def __str__(self) -> str:
        """
        Convert variable to string by putting name and id together.
        
        Returns
        -------
        str
            The name of the variable with its id.
        """
        if self.type == 'template_var' or self.type == 'program_var':
            return self.name
        if self.name is not None:
            return self.name + f'_{self.id}'
        return str(f'var_{self.id}[{self.name}]')

    def __lt__(self, other: UnknownVariable) -> bool:
        """
        Compare variables based on their id (less than).
        
        Parameters
        ----------
        other : UnknownVariable
            The variable to compare with.
            
        Returns
        -------
        bool
            True if the id of the variable is less than the id of the other variable.        
        """
        return self.id < other.id

    def __eq__(self, other: UnknownVariable) -> bool:
        """
        Compare variables based on their id (equal).
        
        Parameters
        ----------
        other : UnknownVariable
            The variable to compare with.
            
        Returns
        -------
        bool
            True if the id of the variable is equal to the id of the other variable.
        """
        return self.id == other.id

    def __hash__(self) -> int:
        """
        Hash variables based on their id.
        
        Returns
        -------
        int
            The hash of the variable.
        """
        return hash(self.id)

    def convert_to_preorder(self) -> str:
        """
        Convert variables into preorder format of a fraction.
        
        Returns
        -------
        str
            The variable in the form of a fraction.
        """
        return f'(/ {self.name}_num {self.name}_den)'



from typing import List

from .Coefficient import Element
from .Constraint import CoefficientConstraint
from .DNF import DNF
from .Polynomial import Coefficient, Monomial, Polynomial, UnknownVariable


class Solver:
    """ 
    This class consist of some static method which are used in other classes
    """

    @staticmethod
    def find_equality_constraint(LHS: Polynomial, RHS: Polynomial) -> List[CoefficientConstraint]:
        """ 
        Given two polynomials that should be equal together, this function 
        finds constraints for the equality. Moreover, all the Coefficient of
        each Monomial in each side should be equal.

        Parameters
        ----------
        LHS : Polynomial
            left hand side of the equality
        RHS : Polynomial
            right hand side of the equality

        Returns
        -------
        List[CoefficientConstraint]
            list of the constraints that needs to be true for equality
        """
        all_degree = set(
            LHS.dict_from_degrees_to_monomials.keys()
        ).union(
            set(RHS.dict_from_degrees_to_monomials.keys())
        )

        all_constraint = []
        for degree in all_degree:
            mono1 = LHS.get_monomial_by_degree(degree)
            mono2 = RHS.get_monomial_by_degree(degree)
            constraint = CoefficientConstraint(
                mono1.coefficient - mono2.coefficient, '=')
            all_constraint.append(constraint)

        return all_constraint

    @staticmethod
    def get_constant_polynomial(variables: List[UnknownVariable], constant) -> Polynomial:
        """
        Generate new polynomial with one monomial and a constant as its coefficient.

        Parameters
        ----------
        variables : List[UnknownVariable]
            Polynomial variables
        constant : str
            The constant of the new polynomial

        Returns
        -------
        Polynomial
            Polynomial with constant as its coefficient
        """
        return Polynomial(variables,
                          [Monomial(
                              variables, [0] * len(variables),
                              Coefficient(
                                  [Element(constant, [])]
                              )
                          )]
                          )

    @staticmethod
    def get_variable_polynomial(variables: List[UnknownVariable], name: str, type_of_var: str = None) -> Polynomial:
        """
        Generate new polynomial with one monomial and a new generated variable as its coefficient

        Parameters
        ----------
        variables : List[UnknownVariable]
            Polynomial variables
        name : str
            Name of the new variable
        type_of_var : str, optional
            Type of the new variable, by default None

        Returns
        -------
        Polynomial
            Polynomial with new variable as its coefficient
        """
        new_variable = UnknownVariable(name=name, type_of_var=type_of_var)
        return Polynomial(variables,
                          [Monomial(
                              variables, [0] * len(variables),
                              Coefficient(
                                  [Element('1', [new_variable])]
                              )
                          )]
                          )

    @staticmethod
    def get_degree_polynomial(variables: List[UnknownVariable], degrees: List[int]) -> Polynomial:
        """ 
        Generate new polynomial with one monomial and given degrees

        Parameters
        ----------
        variables : List[UnknownVariable]
            Polynomial variables
        degrees : List[int]
            Degree of the monomial

        Returns
        -------
        Polynomial
            Polynomial with given degree set
        """
        return Polynomial(variables,
                          [Monomial(
                              variables, degrees,
                              Coefficient(
                                  [Element('1', [])]
                              )
                          )]
                          )

    @staticmethod
    def convert_constraints_to_smt_format(all_constraint: List[DNF], precondition: List[List[DNF]], names: List[str] = None) -> str:
        """ 
        Generate string for declaring constraint in smt format

        Parameters
        ----------
        all_constraint : List[DNF]
            Constraints that should be converted to smt format
        precondition : List[List[DNF]]
            List of precondition that should be added to the constraints
        names : List[str], optional
            List of name for each constraint, by default None

        Returns
        -------
        str
            SMT string format of constraints
        """
        smt_string = ''
        for i, constraint in enumerate(all_constraint):
            if names is None:
                smt_string = smt_string + \
                    f'(assert  {constraint.convert_to_preorder()} )\n'
            else:
                smt_string = smt_string + \
                    f'(assert ( ! {constraint.convert_to_preorder()} :named {names[i]}))\n'

        for constraint in precondition:
            if len(constraint) == 1:
                smt_string = smt_string + \
                    f'(assert  {constraint[0].convert_to_preorder()} )\n'
            elif len(constraint) == 2:
                smt_string = smt_string + \
                    f'(assert (=> {constraint[0].convert_to_preorder()} {constraint[1].convert_to_preorder()}))\n'
        return smt_string

    @staticmethod
    def smt_declare_variable_phase(all_constraint: List[DNF], real: bool = True,
                                   pre_variables: List[UnknownVariable] = []) -> str:
        """
        Generate string format for declaring the variables in smt format.

        Parameters
        ----------
        all_constraint : List[DNF]
            Constraint that their variable should be generated
        real : bool, optional
            Variables should be declared as integer or real valued, by default True
        pre_variables : List[UnknownVariable], optional
            List of variables that should be defined but might not be in constraints, by default []

        Returns
        -------
        str
            SMT string format of declaration phase
        """
        all_variables = Solver.get_all_variable(all_constraint, pre_variables)

        smt_string = ''

        for var in all_variables:
            if real:
                smt_string = smt_string + f'(declare-const {var} Real)\n'
            else:
                smt_string = smt_string + f'(declare-const {var} Int)\n'

        return smt_string

    @staticmethod
    def get_all_variable(all_constraints: List[DNF], pre_variables: List[UnknownVariable] = []) -> List[UnknownVariable]:
        """ 
        Find all the variables in a list of constraint.

        Parameters
        ----------
        all_constraints : List[DNF]
            List of constraints
        pre_variables : List[UnknownVariable], optional
            List of variables that should be defined but might not be in constraints, by default []

        Returns
        -------
        List[UnknownVariable]
            List of variables used in constraints
        """
        all_variables = set(pre_variables)
        for dnf in all_constraints:
            for literal in dnf.literals:
                for constraint in literal:
                    for element in constraint.coefficient.elements:
                        all_variables = all_variables.union(
                            set([var for var in element.variables]))

        return all_variables



from __future__ import annotations

from .Coefficient import Coefficient
from .Polynomial import Polynomial


class PolynomialConstraint:
    """ 
    A class that represent the constraint with a Polynomial on one side and
    zero on the other side.

    Example
    -------
    The polynomial constraint 2x + 3y >= 0 can be represented by this class.

    Parameters
    ----------
    polynomial : Polynomial
        The Polynomial represented in constraint.
    sign : str
        The relation between Polynomial and zero.

    Attributes
    ----------
    polynomial : Polynomial
        The Polynomial represented in constraint.
    sign : str
        The relation between Polynomial and zero.
    """

    def __init__(self, polynomial: Polynomial, sign: str):
        self.polynomial = polynomial
        self.sign = sign
        if sign == '<':
            self.sign = '>'
            self.polynomial = -self.polynomial
        if sign == '<=':
            self.sign = '>='
            self.polynomial = -self.polynomial

    def __str__(self) -> str:
        """
        Convert Constraint to string.

        Returns
        -------
        str
            String format of the PolynomialConstraint.
        """
        return str(self.polynomial) + self.sign + "0"

    def is_strict(self) -> bool:
        """
        Determine if the constraint is strict.

        Returns
        -------
        bool
            True if the constraint is strict.
        """
        return self.sign == '>'

    def __neg__(self) -> PolynomialConstraint:
        """
        Negate a constraint by negating the polynomial and reversing the sign.

        Returns
        -------
        PolynomialConstraint
            New constraint that represent the negation of the previous constraint.
        """
        if self.sign == '>':
            return PolynomialConstraint(-self.polynomial, '>=')
        elif self.sign == '>=':
            return PolynomialConstraint(-self.polynomial, '>')
        elif self.sign == '=':
            return PolynomialConstraint(self.polynomial, '!=')
        elif self.sign == '!=':
            return PolynomialConstraint(self.polynomial, '=')

    def convert_to_preorder(self) -> str:
        """
        Convert constraint to preorder format.

        Returns
        -------
        str
            String in preorder format of the PolynomialConstraint.
        """
        return '(' + self.sign + ' ' + self.polynomial.convert_to_preorder() + ' ' + '0)'


class CoefficientConstraint:
    """
    A class that represent the constraint with a Coefficient on one side and
    zero on the other side.

    Parameters
    ----------
    coefficient : Coefficient
        The Coefficient represented in constraint.
    sign : str
        The relation between Coefficient and zero.

    Attributes
    ----------
    coefficient : Coefficient
        The Coefficient represented in constraint.
    sign : str
        The relation between Coefficient and zero.
    """

    def __init__(self, coefficient: Coefficient, sign: str):
        self.coefficient = coefficient
        self.sign = sign

    def __str__(self) -> str:
        """
        Convert Constraint to string.

        Returns
        -------
        str
            String format of the CoefficientConstraint.
        """
        return str(self.coefficient) + self.sign + "0"

    def is_strict(self) -> bool:
        """
        Determine if the constraint is strict.

        Returns
        -------
        bool
            True if the constraint is strict.
        """
        return self.sign == '>'

    def convert_to_preorder(self) -> str:
        """ 
        Convert constraint to preorder format.

        Returns
        -------
        str
            String in preorder format of the CoefficientConstraint.
        """
        return '(' + self.sign + ' ' + self.coefficient.convert_to_preorder() + ' ' + '0)'

    def is_equality(self) -> bool:
        """
        Determine if the constraint is in form of a variable equal to constant.

        Returns
        -------
        bool
            True if the constraint is in form of a variable equal to constant.
        """
        if self.sign != '=' or len(self.coefficient.elements) > 2:
            return False
        if len(self.coefficient.elements) == 2:
            if len(self.coefficient.elements[0].variables) + len(self.coefficient.elements[1].variables) == 1:
                return True
            else:
                return False
        if len(self.coefficient.elements) == 1:
            if len(self.coefficient.elements[0].variables) == 1:
                return True
            else:
                return False

    def __neg__(self) -> CoefficientConstraint:
        """
        Negate a constraint by negating the coefficient and reversing the sign.

        Returns
        -------
        CoefficientConstraint
            New constraint that represent the negation of the previous constraint.
        """
        if self.sign == '>':
            return CoefficientConstraint(-self.coefficient, '>=')
        elif self.sign == '>=':
            return CoefficientConstraint(-self.coefficient, '>')
        elif self.sign == '=':
            return CoefficientConstraint(self.coefficient, '!=')
        elif self.sign == '!=':
            return CoefficientConstraint(self.coefficient, '=')



from __future__ import annotations

from typing import List

import numpy as np

from .Coefficient import Coefficient
from .UnknownVariable import UnknownVariable


class Monomial:
    """
    A class that represent a Monomial.

    It consists a set of variables with corresponding degree in an array and a Coefficient.

    Example
    -------
    For example `(2*a*b + 1/2*c*c*c + 3.5)*x^2*y^3` is a Monomial.

    Parameters
    ----------
    variables : List[UnknownVariable]
        The sorted list of variables of the Monomial.
    degrees : List[int]
        Degree of each variable.
    coefficient : Coefficient
        The Coefficient of the Monomial

    Attributes
    ----------
    variables : List[UnknownVariable]
        The sorted list of variables of the Monomial.
    degrees : List[int]
        Degree of each variable.
    coefficient : Coefficient
        The Coefficient of the Monomial
    """

    def __init__(self, variables: List[UnknownVariable], degrees: List[int], coefficient: Coefficient):
        variables.sort()
        self.variables = variables
        self.degrees = degrees
        self.coefficient = coefficient

    def __str__(self) -> str:
        """
        Convert Monomial to string.

        Returns
        -------
        str
            String format of the class.
        """
        return '*'.join(['(' + str(self.coefficient) + ')'] +
                        [str(self.variables[i]) + '^' + str(self.degrees[i]) for i in range(len(self.degrees))])

    def __mul__(self, other: Monomial) -> Monomial:
        """
        Multiply two Monomials

        The result is a Monomial which the degree of each variable is sum of
        degrees in multiplicands and the coefficient is the multiply of the two
        coefficient in multiplicands.

        Parameters
        ----------
        other : Monomial
            The other Monomial that should be multiplied

        Returns
        -------
        Monomial
            New Monomial that is the result of multiplication of two Coefficient
        """
        return Monomial(self.variables, np.array(self.degrees) + np.array(other.degrees),
                        self.coefficient * other.coefficient)

    def __eq__(self, other: Monomial) -> bool:
        """
        Compare two Monomials (equality)

        The comparison is based on their length of set of variable and if that
        is equal and all variables and degree are equal they are equal.

        Parameters
        ----------
        other : Monomial
            The other Monomial that should be compared to.

        Returns
        -------
        bool
            Boolean that determine are they equal or not.
        """
        if len(self.variables) != len(other.variables):
            return False
        for i in range(len(self.variables)):
            if self.variables[i] != other.variables[i] or self.degrees[i] != other.degrees[i]:
                return False
        return True

    def __lt__(self, other: Monomial) -> bool:
        """
        Compare two Monomials (less than)

        The comparison is based on their length of set of variable and if that is equal
        based on degree in lexicographical order.

        Parameters
        ----------
        other : Monomial
            The other Monomial that should be compared to.

        Returns
        -------
        bool
            Boolean that determine which one is less than the other.
        """
        if len(self.variables) != len(other.variables):
            return len(self.variables) < len(other.variables)
        for i in range(len(self.variables)):
            if self.degrees[i] != other.degrees[i]:
                return self.degrees[i] < other.degrees[i]
        return True

    def __neg__(self) -> Monomial:
        """
        Negate a Monomial

        For negating a Monomial it is sufficient to just negate the Coefficient.

        Returns
        -------
        Monomial
            A Monomial that is the negated form of the Monomial.
        """
        return Monomial(self.variables, self.degrees, -self.coefficient)

    def is_mono(self) -> bool:
        """
        Determine if the Monomial is a monomial.

        Returns
        -------
        bool
            True if the Monomial is a monomial.
        """
        number_of_nonzero = 0
        for deg in self.degrees:
            if deg != 0 and deg != 1:
                return False
            number_of_nonzero += deg
        return (number_of_nonzero <= 1)

    def get_deg(self) -> int:
        """
        Determine the degree of Monomial.

        Returns
        -------
        bool
            The degree of Monomial
        """
        return np.sum(self.degrees)

    def convert_to_preorder(self) -> str:
        """
        Convert Monomial to preorder format.

        Returns
        -------
        str
            String in preorder format of the class.
        """
        preorder = '( * 1 '
        preorder += self.coefficient.convert_to_preorder()

        for i, var in enumerate(self.variables):
            for _ in range(self.degrees[i]):
                preorder += str(var)
        preorder += ' )'
        return preorder


class Polynomial:
    """
    A class that represent the Polynomial.

    It consists a set of Monomials which should be added together to form the
    Polynomial.

    Example
    -------
    For example `(2*a*b + 1/2*c*c*c + 3.5)*x^2*y^3 + (3 + 2*a*b)*x^0*y^1` is a
    Polynomial.

    Parameters
    ----------
    variables : List[UnknownVariable]
        The sorted list of variables of the Polynomial.
    monomials : List[Monomial]
        The sorted list of Monomials that should be added together.

    Attributes
    ----------
    variables : List[UnknownVariable]
        The sorted list of variables of the Polynomial.
    monomials : List[Monomial]
        The sorted list of Monomials that should be added together.
    dict_from_degrees_to_monomials : dictionary
        A dictionary that maps each degree to corresponding Monomial
    """

    def __init__(self, variables: List[UnknownVariable], monomials: List[Monomial]):
        variables.sort()
        monomials.sort()
        self.variables = variables
        self.monomials = monomials
        self.dict_from_degrees_to_monomials = {}
        for monomial in self.monomials:
            self.dict_from_degrees_to_monomials[tuple(
                monomial.degrees)] = monomial

    def get_monomial_by_degree(self, degree: List[int]) -> Monomial:
        """
        Get a Monomial by its degree.

        Parameters
        ----------
        degree : List[int]
            The degree of the Monomial.

        Returns
        -------
        Monomial
            The Monomial that has the same degree.
        """
        if degree in self.dict_from_degrees_to_monomials.keys():
            return self.dict_from_degrees_to_monomials[degree]
        return Monomial(self.variables, [0] * len(self.variables), Coefficient([]))

    def __str__(self) -> str:
        """
        Convert Polynomial to string.

        Returns
        -------
        str
            String format of the class.
        """
        if len(self.monomials) == 0:
            return '0'
        return '+'.join([str(monomial) for monomial in self.monomials])

    def __add__(self, other: Polynomial) -> Polynomial:
        """
        Sum of two Polynomials

        Sum of two Polynomial is a union of their Monomials.

        Parameters
        ----------
        other : Polynomial
            The other Polynomial that should be added.

        Returns
        -------
        Polynomial
            New Polynomial that is sum of two Coefficient
        """
        return Polynomial(self.variables, self.monomials + other.monomials).revise()

    def __neg__(self) -> Polynomial:
        """
        Negate a Polynomial

        For negating a Polynomial it is sufficient to negate all its Monomial.

        Returns
        -------
        Polynomial
            A Polynomial that is the negated form of the Polynomial.
        """
        monomials = [-monomial for monomial in self.monomials]
        return Polynomial(self.variables, monomials)

    def __sub__(self, other: Polynomial) -> Polynomial:
        """
        Subtract two Polynomial

        Subtract of two Polynomial is adding one with the negated of the other.

        Parameters
        ----------
        other : Polynomial
            The other Polynomial that should be subtracted.

        Returns
        -------
        Polynomial
            New Polynomial that is subtracted of two Polynomial
        """
        return self + (-other)

    def __mul__(self, other: Polynomial) -> Polynomial:
        """
        Multiply two Polynomial

        The result is a Polynomial which consist of Monomials that are equal to
        the multiply of two Monomial from each Polynomial. Numpy is used for
        that.

        Parameters
        ----------
        other : Polynomial
            The other Polynomial that should be multiplied

        Returns
        -------
        Polynomial
            New Polynomial that is multiplied of two Polynomial
        """
        return Polynomial(self.variables,
                          np.array(
                              np.matmul(
                                  np.array(self.monomials).reshape(1, -1).T,
                                  np.array(other.monomials).reshape(1, -1)
                              )
                          ).reshape(1, -1)[0]
                          ).revise()

    def revise(self) -> Polynomial:
        """
        Revise the Polynomial

        It means that the Monomials that are the same added together.

        Example
        -------
        For example `(3*a)*x^1 + (2*a*b + 1*a)*x^1`  is a Polynomial and after
        revise it returns `(3*a + 2*a*b + 1*a)*x^1`.

        Returns
        -------
        Polynomial
            The revised format of the Polynomial.
        """
        self.monomials.sort()
        new_list = []
        i = 0
        while i < len(self.monomials):
            coefficient = Coefficient([])
            j = i
            while j < len(self.monomials):
                if self.monomials[i] == self.monomials[j]:
                    coefficient += self.monomials[j].coefficient
                else:
                    break
                j += 1
            new_list.append(
                Monomial(self.monomials[i].variables, self.monomials[i].degrees, coefficient))
            i = j

        return Polynomial(self.variables, new_list)

    def add_variables(self, new_variables: List[UnknownVariable]) -> Polynomial:
        """
        This function adds a set of new variable to each Monomial of a
        Polynomial.

        Parameters
        ----------
        new_variables : List[UnknownVariable]
            The list of the new variables that should add to the Polynomial.

        Returns
        -------
        Polynomial
            New Polynomial with the new list of the variables.
        """
        monomials = []
        for monomial in self.monomials:
            monomials.append(Monomial(monomial.variables + new_variables,
                                      monomial.degrees +
                                      [0] * len(new_variables),
                                      monomial.coefficient)
                             )
        return Polynomial(self.variables + new_variables, monomials)

    def is_linear(self) -> bool:
        """
        Determine if the Polynomial is linear.

        Returns
        -------
        bool
            True if the Polynomial is linear.
        """
        for mono in self.monomials:
            if not mono.is_mono():
                return False
        return True

    def get_deg(self) -> int:
        """
        Determine degree of Polynomial.

        Returns
        -------
        int
            degree of Polynomial.
        """
        deg = 0
        for mono in self.monomials:
            deg = max(deg, mono.get_deg())
        return deg
    def convert_to_preorder(self) -> str:
        """
        Convert Polynomial to preorder format.

        Returns
        -------
        str
            String in preorder format of the class.
        """
        if len(self.monomials) == 0:
            return '0'
        if len(self.monomials) == 1:
            return self.monomials[0].convert_to_preorder()
        preorder = '( + 0 '
        for i in range(0, len(self.monomials)):
            preorder += self.monomials[i].convert_to_preorder() + ' '
        preorder += ' )'
        return preorder



from src.polyqent.main import add_default_config, execute, load_config


def get_test_solver_instance():
    """
    Utility function to create a solver instance with the same
    constraints as in `tests/res/test_instance.smt2`
    """
    from pysmt.shortcuts import (GE, GT, LE, And, ForAll, Implies, Plus, Real,
                                 Solver, Symbol, Times)
    from pysmt.typing import REAL

    c_1 = Symbol("c_1", REAL)
    c_2 = Symbol("c_2", REAL)
    c_3 = Symbol("c_3", REAL)
    c_4 = Symbol("c_4", REAL)

    x = Symbol("x", REAL)

    solver = Solver(name="z3")

    # (assert (> (+ c_3 c_4) 0))
    solver.add_assertion(
        GT(Plus(c_3, c_4), Real(0))
    )

    # (assert (forall ((x Real) ) (=> (and (>= x -1024) (>= 1023 x)) (>= (+ (* c_1 x) c_2) 0) )))
    solver.add_assertion(
        ForAll([x], Implies(And(GE(x, Real(-1024)), GE(Real(1023), x)),
               GE(Plus(Times(c_1, x), c_2), Real(0))))
    )

    # (assert (forall ((x Real) ) (=> (and (>= x -1024) (>= 1023 x) (>= x 1)) (and (>= (+ (* c_1 (+ x -1)) c_2) 0)  (<= (+ (* c_1 (+ x -1)) c_2) (+ (* c_1 x) c_2 -1) ) ))))
    solver.add_assertion(
        ForAll([x], Implies(And(GE(x, Real(-1024)), GE(Real(1023), x), GE(x, Real(1))),
                            And(GE(Plus(Times(c_1, Plus(x, Real(-1))), c_2), Real(0)),
                                LE(Plus(Times(c_1, Plus(x, Real(-1))), c_2), Plus(Times(c_1, x), c_2, Real(-1))))))
    )

    # (assert (forall ((x Real) ) (=> (and (>= x -1024) (>= 1023 x) (<= x 0)) (and (>= (+ (* c_3 x) c_4) 0) (<= (+ (* c_3 x) c_4 ) (+ (* c_1 x) c_2 -1) ) ))))
    solver.add_assertion(
        ForAll([x], Implies(And(GE(x, Real(-1024)), GE(Real(1023), x), LE(x, Real(0))),
                            And(GE(Plus(Times(c_3, x), c_4), Real(0)),
                                LE(Plus(Times(c_3, x), c_4), Plus(Times(c_1, x), c_2, Real(-1))))))
    )

    return solver


def get_test_config_dict():
    """
    Utility function to create a configuration dictionary with the same
    configuration as in `tests/res/test_config.json`
    """
    return {
        "theorem_name": "handelman",
        "degree_of_sat": 2,
        "degree_of_nonstrict_unsat": 0,
        "degree_of_strict_unsat": 1,
        "max_d_of_strict": 4,
        "solver_name": "z3",
        "output_path": "fully_existentially_constraints.txt",
        "unsat_core_heuristic": False,
        "SAT_heuristic": True,
        "integer_arithmetic": False
    }


def test_add_default_config_on_empty_dict():

    assert add_default_config({}) == {
        "SAT_heuristic": False,
        "degree_of_sat": 0,
        "degree_of_nonstrict_unsat": 0,
        "degree_of_strict_unsat": 0,
        "max_d_of_strict": 0,
        "unsat_core_heuristic": False,
        "integer_arithmetic": False
    }


def test_add_default_config_on_partial_dict():

    assert add_default_config({"SAT_heuristic": True}) == {
        "SAT_heuristic": True,
        "degree_of_sat": 0,
        "degree_of_nonstrict_unsat": 0,
        "degree_of_strict_unsat": 0,
        "max_d_of_strict": 0,
        "unsat_core_heuristic": False,
        "integer_arithmetic": False
    }


def test_add_default_config_on_dict_with_additional_content():

    assert add_default_config({
        "theorem_name": "farkas",
        "SAT_heuristic": True,
        "degree_of_sat": 1,
        "degree_of_nonstrict_unsat": 2,
        "degree_of_strict_unsat": 3,
        "max_d_of_strict": 4,
        "unsat_core_heuristic": True,
        "integer_arithmetic": True
    }) == {
        "theorem_name": "farkas",
        "SAT_heuristic": True,
        "degree_of_sat": 1,
        "degree_of_nonstrict_unsat": 2,
        "degree_of_strict_unsat": 3,
        "max_d_of_strict": 4,
        "unsat_core_heuristic": True,
        "integer_arithmetic": True
    }


def test_load_config():

    assert load_config("tests/res/test_config.json") == get_test_config_dict()


def test_execute_with_smt_path_and_config_path():

    sat, model = execute("tests/res/test_instance.smt2",
                         "tests/res/test_config.json")
    assert sat == "sat"
    assert model == {
        'c_1': '1.0', 'c_2': '(/ 2051.0 2.0)', 'c_3': '0.0', 'c_4': '(/ 1.0 2.0)'}


def test_execute_with_solver_and_config_path():

    sat, model = execute(get_test_solver_instance(),
                         "tests/res/test_config.json")
    assert sat == "sat"
    assert model == {
        'c_1': '1.0', 'c_2': '(/ 2051.0 2.0)', 'c_3': '0.0', 'c_4': '(/ 1.0 2.0)'}


def test_execute_with_smt_path_and_config_dict():

    sat, model = execute("tests/res/test_instance.smt2",
                         get_test_config_dict())
    assert sat == "sat"
    assert model == {
        'c_1': '1.0', 'c_2': '(/ 2051.0 2.0)', 'c_3': '0.0', 'c_4': '(/ 1.0 2.0)'}


def test_execute_with_solver_and_config_dict():

    sat, model = execute(get_test_solver_instance(),
                         get_test_config_dict())
    assert sat == "sat"
    assert model == {
        'c_1': '1.0', 'c_2': '(/ 2051.0 2.0)', 'c_3': '0.0', 'c_4': '(/ 1.0 2.0)'}



from typing import List, Tuple

from .Constraint import CoefficientConstraint, PolynomialConstraint
from .Polynomial import Polynomial
from .Solver import Solver
from .UnknownVariable import UnknownVariable


class Handelman:
    """ 
    A class that performs the constraint transformation according to Handelman
    Theorem.    

    Parameters
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm
    max_d_for_sat : int, optional
        maximum degree of monoids when finding sat constraints, by default 0
    max_d_for_unsat : int, optional
        maximum degree of monoids when finding unsat constraints, by default 0

    Attributes
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm
    max_d_for_sat : int
        maximum degree of monoids when finding sat constraints
    max_d_for_unsat : int
        maximum degree of monoids when finding unsat constraints
    """

    def __init__(self, variables: List[UnknownVariable], LHS: List[PolynomialConstraint], RHS: PolynomialConstraint,
                 max_d_for_sat: int = 0, max_d_for_unsat: int = 0):
        self.variables = variables
        self.RHS = RHS
        self.LHS = LHS
        self.max_d_for_sat = max_d_for_sat
        self.max_d_for_unsat = max_d_for_unsat

    def get_lists_with_fixed_len(self, lst: List[int], n: int, max_d: int) -> List[List[int]]:
        """
        This function generates all the list of length n with elements in range
        0 to max_d and sum of all elements is less than or equal to max_d.

        Parameters
        ----------
        lst : List[int]
            The list that is generated so far.
        n : int
            The length of the list.
        max_d : int
            The maximum value of each element in the list.

        Returns
        -------
        List[List[int]]
            List of all the list of length n with elements in range 0 to max_d
            and sum of all elements is less than or equal to max_d.
        """
        if len(lst) == n:
            return [lst]
        ans = []
        for i in range(max_d - sum(lst)+1):
            ans = ans + self.get_lists_with_fixed_len(lst + [i], n, max_d)
        return ans

    def get_monoids(self, max_d: int) -> Tuple[List[Polynomial], List[bool]]:
        """ 
        This function creates monoids of the given degree of all left hand side
        polynomials.

        Parameters
        ----------
        max_d : int
            Maximum degree of each monoid

        Returns
        -------
        Tuple[List[Polynomial], List[bool]]
            List of monoids and a list of boolean which the ith element is true
            if the ith monoid is strict.
        """
        monoids = []
        is_strict = []
        all_degrees = self.get_lists_with_fixed_len([], len(self.LHS), max_d)
        for degree_of_each_lhs in all_degrees:
            poly = Solver.get_constant_polynomial(self.variables, '1')
            is_the_new_monoid_strict = True

            for i in range(len(self.LHS)):
                if (not self.LHS[i].is_strict()) and (degree_of_each_lhs[i] > 0):
                    is_the_new_monoid_strict = False
                for d in range(degree_of_each_lhs[i]):
                    poly = poly * self.LHS[i].polynomial

            is_strict.append(is_the_new_monoid_strict)
            monoids.append(poly)

        return monoids, is_strict

    def get_poly_sum(self, max_d: int, need_strict: bool = False) -> Tuple[Polynomial, CoefficientConstraint]:
        """ 
        This function returns a polynomial `y_0 + y_1*f_1 + y_2*f_2 ...+ y_n*f_n`
        where `f_i` are monoids of left hand side polynomials and `y_i` are newly
        created variable in handelman theorem.

        Parameters
        ----------
        max_d : int
            Maximum degree of each monoid
        need_strict : bool, optional
            Is it generated for the 0 > 0 case or not, by default False

        Returns
        -------
        Tuple[Polynomial, CoefficientConstraint]
            Polynomial of sum of all monoids of left hand side with new variables
            and corresponding constraints.
        """
        polynomial_of_sum = Polynomial(self.variables, [])

        all_monoid, is_strict = self.get_monoids(max_d)
        constraints = []
        sum_of_strict = Solver.get_constant_polynomial(self.variables, '0')

        if self.RHS.is_strict():
            new_var = Solver.get_variable_polynomial(
                self.variables, 'y0', 'generated_for_handelman_in_strict_case')
            polynomial_of_sum = polynomial_of_sum + new_var
            constraints.append(CoefficientConstraint(
                new_var.monomials[0].coefficient, '>='))
            sum_of_strict = sum_of_strict + new_var

        for i, monoid in enumerate(all_monoid):
            new_var_poly = Solver.get_variable_polynomial(self.variables, f'y{i + 1}',
                                                          'generated_for_Handelman')
            polynomial_of_sum = polynomial_of_sum + (new_var_poly * monoid)
            constraints.append(CoefficientConstraint(
                new_var_poly.monomials[0].coefficient, '>='))

            if is_strict[i]:
                sum_of_strict = sum_of_strict + new_var_poly

        if need_strict or self.RHS.is_strict():
            constraints.append(CoefficientConstraint(
                sum_of_strict.monomials[0].coefficient, '>'))
        return polynomial_of_sum, constraints

    def get_SAT_constraint(self) -> List[CoefficientConstraint]:
        """
        A function to find the constraints when the LHS => RHS is satisfiable.

        Returns
        -------
        List[CoefficientConstraint]
            List of coefficient constraints when it is satisfiable.
        """
        polynomial_of_sum, constraints = self.get_poly_sum(self.max_d_for_sat)
        return Solver.find_equality_constraint(polynomial_of_sum, self.RHS.polynomial) + constraints

    def get_UNSAT_constraint(self, need_strict: bool = False) -> List[CoefficientConstraint]:
        """ 
        A function to find the constraints when it is not satisfiable.

        Two set of constraint can be generated:
            1. `LHS => -1 >= 0`
            2. `LHS => 0 > 0`

        Parameters
        ----------
        need_strict : bool, optional
            Determine which set of constraint to generate, by default False

        Returns
        -------
        List[CoefficientConstraint]
            List of coefficient constraints when it is not satisfiable.
        """
        if need_strict:
            polynomial_of_sum, constraints = self.get_poly_sum(
                self.max_d_for_unsat, need_strict)
            return Solver.find_equality_constraint(polynomial_of_sum,
                                                   Polynomial(self.variables, [])) + constraints

        polynomial_of_sum, constraints = self.get_poly_sum(
            self.max_d_for_unsat)
        return Solver.find_equality_constraint(polynomial_of_sum,
                                               Solver.get_constant_polynomial(self.variables, '-1')) + constraints



from __future__ import annotations

from fractions import Fraction
from typing import List, Union

import numpy as np

from .UnknownVariable import UnknownVariable


class Element:
    """
    A class that represent an element.

    It consists a rational constant and a set of variables which should be
    multiplied together to form the element.

    Examples
    -------
    - The term `2*a*b` is an element where 2 is the constant and a, b are variables.
        
        >>> Element(2, [UnknownVariable('a'), UnknownVariable('b')])
        
    - The term `1/2*c*c*c` is an element where 1/2 is the constant and c is the variable.
        
        >>> Element(1/2, [UnknownVariable('c'), UnknownVariable('c'), UnknownVariable('c')])


    Parameters
    ----------
    constant : Union[float, str, Fraction]
        It is the constant to multiply with other variables
    variables : List[UnknownVariable], optional
        The sorted list of variables that should be multiplied together, by default []


    Attributes
    ----------
    constant : Fraction
        The constant of the element
    variables : List[UnknownVariable]
        The sorted list of variables that should be multiplied together
    """

    def __init__(self, constant: Union[float, str, Fraction], variables: List[UnknownVariable] = []):
        self.constant = Fraction(constant)
        variables.sort()
        self.variables = variables

    def __str__(self) -> str:
        """
        Convert Element to string.

        Returns
        -------
        str
            String representation of the Element.
        """
        return '*'.join([str(self.constant)] + [str(var) for var in self.variables])

    def __mul__(self, other: Element) -> Element:
        """
        Multiply two elements.

        Parameters
        ----------
        other : Element
            The other factor for multiplication

        Returns
        -------
        Element
            New element that is the product of two elements
        """
        return Element((self.constant * other.constant), (self.variables + other.variables))

    def __add__(self, other: Element) -> Element:
        """
        Add two elements.

        Parameters
        ----------
        other : Element
            The other element that should be added

        Returns
        -------
        Element
            New element that is the sum of two elements
        """
        return Coefficient([self, other])

    def __lt__(self, other: Element) -> bool:
        """
        Compare two elements (less than).

        The comparison is based on their length of set of variable and if that
        is equal based on the set of variable lexicographically and if that is
        equal too based on their constant.

        Parameters
        ----------
        other : Element
            The other element that should be compared to.

        Returns
        -------
        bool
            Boolean that determine which one is less than the other.
        """
        if len(self.variables) == len(other.variables):
            for i in range(len(self.variables)):
                if not (self.variables[i] == other.variables[i]):
                    return self.variables[i] < other.variables[i]
            return self.constant < other.constant
        else:
            return len(self.variables) < len(other.variables)

    def __eq__(self, other: Element) -> bool:
        """
        Compare two elements (equal).

        The comparison is based on their length of set of variable and if that
        is equal and all variables are equal they are equal.      

        Parameters
        ----------
        other : Element
            The other element that should be compared to.

        Returns
        -------
        bool
            Boolean that determine are they equal or not.
        """

        if len(self.variables) == len(other.variables):
            for i in range(len(self.variables)):
                if not (self.variables[i] == other.variables[i]):
                    return False
            return True
        else:
            return False

    def __neg__(self) -> Element:
        """
        Negate an element.

        For negating an element it is sufficient to just negate the constant.

        Returns
        -------
        Element
            A new element that is the negated form of the element.
        """
        return Element(-self.constant, self.variables)

    def convert_to_preorder(self) -> str:
        """
        Convert Element to preorder format string.

        Returns
        -------
        str
            String in preorder format of the class
        """
        if self.constant == 0:
            return '0'
        if self.constant.denominator == 1:
            preorder = str(self.constant.numerator)
            if self.constant < 0:
                preorder = f'(- {-self.constant.numerator})'
        else:
            preorder = f'(/ {self.constant.numerator} {self.constant.denominator})'
            if self.constant < 0:
                preorder = f'(- (/ {-self.constant.numerator} {self.constant.denominator}))'
        if len(self.variables) == 0:
            return preorder
        preorder = '(* 1 ' + preorder + ' '
        for var in self.variables:
            preorder += str(var) + ' '
        preorder += ' )'
        return preorder


class Coefficient:
    """
    A class that represent a coefficient.

    It consists a list of elements that should be added together to form the coefficient.

    Examples
    --------
    For example `2*a*b + 1/2*c*c*c + 3.5` is a coefficient.

    >>> Coefficient(
    ...     [
    ...         Element(2, [UnknownVariable('a'), UnknownVariable('b')]),
    ...         Element(1/2, [UnknownVariable('c'), UnknownVariable('c'), UnknownVariable('c')]),
    ...         Element(3.5)
    ...     ])
    

    Parameters
    ----------
    elements : List[Element], optional
        The list of elements that should be added together, by default []

    Attributes
    ----------
    elements : List[Element]
        The list of elements that should be added together    
    """

    def __init__(self, elements: List[Element] = []):
        elements.sort()
        self.elements = elements

    def __str__(self) -> str:
        """
        Convert Coefficient to string.

        Returns
        -------
        str
            String representation of the Coefficient.
        """
        if len(self.elements) == 0:
            return '0'
        return '+'.join([str(element) for element in self.elements])

    def __mul__(self, other: Coefficient) -> Coefficient:
        """
        Multiply two Coefficient.

        Parameters
        ----------
        other : Coefficient
            The other factor for multiplication

        Returns
        -------
        Coefficient
            New coefficient that is the product of two coefficients
        """
        return Coefficient(
            np.array(
                np.matmul(
                    np.array(self.elements).reshape(1, -1).T,
                    np.array(other.elements).reshape(1, -1)
                )
            ).reshape(1, -1)[0]
        ).revise()

    def __add__(self, other: Union[Coefficient, Element]) -> Coefficient:
        """
        Add two Coefficient.

        Parameters
        ----------
        other : Union[Coefficient, Element]
            The other element that should be added to the Coefficient

        Returns
        -------
        Coefficient
            New coefficient that is the sum of two coefficients
        """
        if type(other) is Coefficient:
            return Coefficient(self.elements + other.elements).revise()
        if type(other) is Element:
            return Coefficient(self.elements + [other]).revise()

    def revise(self) -> Coefficient:
        """ 
        Revise the Coefficient.

        It sorts the elements and combines the elements with the same variables.

        Examples
        --------
        For example `3*a + 2*a*b + 1*a` is a Coefficient and after revise it returns `4*a + 2*a*b`.

        Returns
        -------
        Coefficient
            The revised format of the Coefficient.
        """
        self.elements.sort()
        new_list = []
        i = 0
        while i < len(self.elements):
            constant = Fraction('0')
            j = i
            while j < len(self.elements):
                if self.elements[i] == self.elements[j]:
                    constant += self.elements[j].constant
                else:
                    break
                j += 1
            new_list.append(Element(constant, self.elements[i].variables))
            i = j

        return Coefficient(new_list)

    def __neg__(self) -> Coefficient:
        """
        Negate a Coefficient.

        For negating a Coefficient it is sufficient to negate each element.

        Returns
        -------
        Coefficient
            A new coefficient that is the negated form of the Coefficient.
        """
        elements = [-element for element in self.elements]
        return Coefficient(elements)

    def __sub__(self, other: Coefficient) -> Coefficient:
        """
        Subtract two Coefficient.

        Subtract of two Coefficient is adding one with the negated of the other.

        Parameters
        ----------
        other : Coefficient
            The other Coefficient that should be subtracted.

        Returns
        -------
        Coefficient
            New coefficient that is subtracted of two Coefficient
        """
        return self + (-other)

    def convert_to_preorder(self) -> str:
        """
        Convert Coefficient to preorder format string.

        Returns
        -------
        str
            String in preorder format of the class
        """
        if len(self.elements) == 0:
            return '0'
        if len(self.elements) == 1:
            return self.elements[0].convert_to_preorder()
        preorder = '(+ 0 '
        not_zero = False
        for i in range(0, len(self.elements)):
            if self.elements[i].constant == 0:
                continue
            else:
                not_zero = True
            preorder += self.elements[i].convert_to_preorder() + ' '
        preorder += ')'
        if not_zero:
            return preorder
        else:
            return '0'



import subprocess

def test_cmd_tool():
    cmd = "./PolyQEnt tests/res/test_instance.smt2 tests/res/test_config.json"
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = process.communicate()
    out, err = out.decode('utf-8'), err.decode('utf-8')

    assert 'The system is sat' in out
    
    model = ['Model:', 'c_1: 1.0', 'c_2: (/ 2051.0 2.0)', 'c_3: 0.0', 'c_4: (/ 1.0 2.0)']
    for line in model:
        assert line in out
    
    assert err == ''
    
    assert process.returncode == 0
    
    
    
test_cmd_tool()


from typing import List, Tuple

import numpy as np

from .Constraint import CoefficientConstraint, PolynomialConstraint
from .Polynomial import Polynomial
from .Solver import Solver
from .UnknownVariable import UnknownVariable


class Putinar:
    """ 
    A class that performs the constraint transformation according to Putinar
    Theorem.

    Parameters
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm
    max_d_for_sat : int, optional
        maximum degree of monoids when finding sat constraints, by default 0
    max_d_for_unsat : int, optional
        maximum degree of monoids when finding unsat constraints, by default 0
    max_d_for_unsat_strict : int, optional
        maximum degree of monoids when finding unsat constraints in strict case, by default 0
    degree_for_new_var : int, optional
        degree of new variable that is generated for strict case in right hand side of equality, by default 0

    Attributes
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm
    max_d_for_sat : int
        maximum degree of monoids when finding sat constraints
    max_d_for_unsat : int
        maximum degree of monoids when finding unsat constraints
    max_d_for_unsat_strict : int
        maximum degree of monoids when finding unsat constraints in strict case
    degree_for_new_var : int
        degree of new variable that is generated for strict case in right hand side of equality
    """

    def __init__(self, variables: List[UnknownVariable], LHS: List[PolynomialConstraint], RHS: PolynomialConstraint,
                 max_d_for_sat: int = 0, max_d_for_unsat: int = 0, max_d_for_unsat_strict: int = 0, degree_for_new_var: int = 0):
        self.variables = variables
        self.RHS = RHS
        self.LHS = LHS
        self.max_d_for_sat = max_d_for_sat
        self.max_d_for_unsat = max_d_for_unsat
        self.max_d_for_unsat_strict = max_d_for_unsat_strict
        self.degree_for_new_var = degree_for_new_var

    def get_lower_triangular_matrix(self, dim: int) -> Tuple[List[List[Polynomial]], List[CoefficientConstraint]]:
        """
        This function generates a lower triangular matrix `L` with positive
        elements in diagonal and returns `L * L^T`.

        Parameters
        ----------
        dim : int
            The dimension of the generated matrix

        Returns
        -------
        List[List[Polynomial]]
            A positive semidefinite matrix.
        List[CoefficientConstraint]
            List of constraints on the diagonal elements of the matrix 
        """
        matrix = np.array(
            [[Polynomial(self.variables, []) for __ in range(dim)]
             for _ in range(dim)]
        )
        constraints = []
        for i in range(dim):
            for j in range(i + 1):
                matrix[i][j] = Solver.get_variable_polynomial(self.variables, f'l_{i}{j}',
                                                              'generated_for_semidefinite_matrix')
            constraints.append(CoefficientConstraint(
                matrix[i][i].monomials[0].coefficient, '>='))

        return np.matmul(matrix, matrix.T), constraints

    @staticmethod
    def get_monoids(variables, max_d: int) -> List[Polynomial]:
        """ 
        This function gives the monoid of a maximum degree on set of variables.

        Parameters
        ----------
        variables : List[UnknownVariable]
            The set of variables
        max_d : int
            Maximum degree of generated monoids

        Returns
        -------
        List[Polynomial]
            List of monoids defined on the set of variable
        """
        all_monoid = []
        for mask in range((max_d + 1) ** len(variables)):
            mask_copy = mask
            degrees = []
            for i in range(len(variables)):
                degrees.append(mask_copy % (max_d + 1))
                mask_copy //= (max_d + 1)
            if sum(degrees) > max_d:
                continue
            all_monoid.append(Solver.get_degree_polynomial(variables, degrees))

        return all_monoid

    def get_sum_of_square(self, max_d: int) -> Tuple[Polynomial, List[CoefficientConstraint]]:
        """ 
        The function generates a new sum-of-square template on the set of class 
        variables and given maximum degree.

        Parameters
        ----------
        max_d : int
            Maximum degree of the generated templates

        Returns
        -------
        Polynomial
            Polynomial represents the sum-of-square template
        List[CoefficientConstraint]
            List of constraints on the coefficients of the template
        """
        monoid = np.array(Putinar.get_monoids(
            self.variables, (max_d // 2))).reshape(1, -1)
        semi_definite_matrix, constraint = self.get_lower_triangular_matrix(
            monoid.shape[1])
        poly_with_more_degree = np.matmul(
            np.matmul(monoid, semi_definite_matrix), monoid.T)[0][0]
        monomials = []
        for i, monomial in enumerate(poly_with_more_degree.monomials):
            new_poly = Solver.get_variable_polynomial(self.variables, f't_{i}', 'eta_generated_for_sum_of_square') \
                * Solver.get_degree_polynomial(self.variables, monomial.degrees)
            constraint.append(
                CoefficientConstraint(
                    new_poly.monomials[0].coefficient - monomial.coefficient, '=')
            )
            monomials.append(new_poly.monomials[0])
        return Polynomial(poly_with_more_degree.variables, monomials), constraint

    def get_poly_sum(self, max_d: int) -> Tuple[Polynomial, List[CoefficientConstraint]]:
        """ 
        This function returns a polynomial `y_0(rhs is struct) + h_0 + h_1*g_1 ... + h_n*g_n`
        where `h_i` are sum-of-square polynomial and `g_i` are left hand side polynomials.

        Parameters
        ----------
        max_d : int
            Maximum degree of the generated templates

        Returns
        -------
        Polynomial
            Polynomial of sum of all left hand side with new variables
        List[CoefficientConstraint]
            List of constraints on the coefficients of the templates
        """
        polynomial_of_sum, all_constraints = self.get_sum_of_square(max_d)
        strict_poly = Solver.get_constant_polynomial(self.variables, 0)
        if self.RHS.is_strict():
            new_var = Solver.get_variable_polynomial(
                self.variables, 'y0', 'generated_for_putinar_in_strict_case')
            strict_poly = strict_poly + new_var
            polynomial_of_sum = polynomial_of_sum + new_var
            all_constraints.append(CoefficientConstraint(
                new_var.monomials[0].coefficient, '>='))
        for i, left_constraint in enumerate(self.LHS):
            left_poly = left_constraint.polynomial
            new_sum_of_square, constraint = self.get_sum_of_square(max_d)
            if self.RHS.is_strict() and left_constraint.is_strict():
                new_var = Solver.get_variable_polynomial(
                    self.variables, 'y0', 'generated_for_putinar_in_strict_case')
                strict_poly = strict_poly + new_var
                new_sum_of_square = new_sum_of_square + new_var

                all_constraints.append(CoefficientConstraint(
                    new_var.monomials[0].coefficient, '>='))
            all_constraints = all_constraints + constraint
            polynomial_of_sum = polynomial_of_sum + \
                (new_sum_of_square * left_poly)
        if self.RHS.is_strict():
            all_constraints.append(CoefficientConstraint(
                strict_poly.monomials[0].coefficient, '>'))
        return polynomial_of_sum, all_constraints

    def get_SAT_constraint(self) -> List[CoefficientConstraint]:
        """ 
        A function to find the constraints when the LHS => RHS is satisfiable

        Returns
        -------
        List[CoefficientConstraint]
            List of coefficient constraints when it is satisfiable
        """
        polynomial_of_sum, constraints = self.get_poly_sum(self.max_d_for_sat)
        return constraints + Solver.find_equality_constraint(polynomial_of_sum, self.RHS.polynomial)

    def get_UNSAT_constraint(self, need_strict=False) -> List[CoefficientConstraint]:
        """ 
        A function to find the constraints when it is not satisfiable.

        Two set of constraint can be generated:
            1. `LHS => -1 >= 0`
            2. `w_j^2*d = h_1*(g_1 - w_1^2) + h_2*(g_2 - w_2^2) ... h_n*(g_n - w_n^2)` for some `j` where `g_j` is strict 

        Parameters
        ----------
        need_strict : bool, optional
            Determine which set of constraint to generate, by default False

        Returns
        -------
        List[CoefficientConstraint]
            List of coefficient constraints when it is not satisfiable
        """
        if need_strict:
            return self.handel_strict(self.max_d_for_unsat_strict, self.degree_for_new_var)
        polynomial_of_sum, constraints = self.get_poly_sum(
            self.max_d_for_unsat)
        return constraints + Solver.find_equality_constraint(polynomial_of_sum,
                                                             Solver.get_constant_polynomial(
                                                                 self.RHS.polynomial.variables, '-1'))

    def handel_strict(self, max_d: int, power_of_new_var: int) -> List[List[CoefficientConstraint]]:
        """ 
        Handles the second form of unsatisfiablity

        Parameters
        ----------
        max_d : int
            Maximum degree of the generated templates
        power_of_new_var : int
            Power of the new variable that is generated for strict case

        Returns
        -------
        List[List[CoefficientConstraint]]
            List of coefficient constraints when it is not satisfiable
        :param max_d: maximum degree for template polynomials
        :param power_of_new_var: power of newly generated variables
        :return: list of constraint for each equality
        """
        new_variables = [UnknownVariable(name=f'w_{i + 1}', type_of_var='generated_for_strict_unsat') for i in
                         range(len(self.LHS))]
        new_set_of_cons = []
        for left_constraint in self.LHS:
            new_set_of_cons.append(PolynomialConstraint(left_constraint.polynomial.add_variables(new_variables), left_constraint.sign
                                                        )
                                   )
        all_variable = self.variables + new_variables
        all_monoids = Putinar.get_monoids(all_variable, max_d)
        all_constraints = []
        for j, left_constraint in enumerate(new_set_of_cons):
            if left_constraint.is_strict():
                poly_sum = Polynomial(all_variable, [])
                for i, g in enumerate(new_set_of_cons):
                    degrees = [0] * (len(all_variable))
                    degrees[len(self.variables) + i] = 1
                    new_var_poly = Solver.get_degree_polynomial(
                        all_variable, degrees)
                    poly_sum = poly_sum + \
                        (Putinar.get_general_template(all_variable, all_monoids) *
                         (g.polynomial - (new_var_poly * new_var_poly)))

                degrees = [0] * (len(all_variable))
                degrees[len(self.variables) + j] = 1
                new_var_poly = Solver.get_degree_polynomial(
                    all_variable, degrees)
                poly_of_var_to_power = Solver.get_constant_polynomial(
                    all_variable, '1')
                for _ in range(2 * power_of_new_var):
                    poly_of_var_to_power = poly_of_var_to_power * new_var_poly
                all_constraints.append(Solver.find_equality_constraint(
                    poly_of_var_to_power, poly_sum))
        return all_constraints

    @staticmethod
    def get_general_template(all_variables: List[UnknownVariable], all_monoids: List[Polynomial]) -> List[Polynomial]:
        """
        Generate a new variable for every monoid and multiply it and return the
        sum of all.

        Parameters
        ----------
        all_variables : List[UnknownVariable]
            Variables of the polynomials
        all_monoids : List[Polynomial]
            Monoids which should be multiplied by new variable

        Returns
        -------
        Polynomial
            A polynomial of the sum of all monoid multiplied by new variables
        """
        new_unknown_variables = [
            Solver.get_variable_polynomial(
                all_variables, name=f'eta{i + 1}', type_of_var='generated_for_template_poly_strict')
            for i in range(len(all_monoids))]
        return np.matmul(np.array(new_unknown_variables).reshape(1, -1), np.array(all_monoids).reshape(-1, 1))[0][0]



from typing import List

import lark
from lark import Lark

from .Polynomial import Monomial, Polynomial
from .Solver import Solver
from .UnknownVariable import UnknownVariable


def find_index_of_variable(name: str, all_variable: List[UnknownVariable]) -> int:
    """
    Find the index of a variable in the list of variables, given the name of
    the variable.

    Parameters
    ----------
    name : str
        The name of the variable.    
    all_variable : List[UnknownVariable]
        List of all variables to search.

    Returns
    -------
    int
        The index of the variable in the list of variables.
    """
    all_variable.sort()
    for i in range(len(all_variable)):
        if name == all_variable[i].name:
            return i
    print(f'no such var declared with name : {name}')
    return -1


def traverse(parse_tree: lark.Tree, all_variables: List[UnknownVariable]) -> Polynomial:
    """
    This function traverse the parse tree generated by lark and return a Polynomial.

    Parameters
    ----------
    parse_tree : lark.Tree
        A node in parse tree.(starting from root)
    all_variables : List[UnknownVariable]
        List of variables in the polynomial.

    Returns
    -------
    Polynomial
        Polynomial that the parse tree represents.
    """
    if parse_tree.data == 'start':
        return traverse(parse_tree.children[0], all_variables)
    elif parse_tree.data == 'expression':
        if len(parse_tree.children) == 1:
            return traverse(parse_tree.children[0], all_variables)
        elif parse_tree.children[1] == '+':
            return traverse(parse_tree.children[0], all_variables) + traverse(parse_tree.children[2], all_variables)
        elif parse_tree.children[1] == '-':
            return traverse(parse_tree.children[0], all_variables) - traverse(parse_tree.children[2], all_variables)

    elif parse_tree.data == 'term':
        if len(parse_tree.children) == 1:
            return traverse(parse_tree.children[0], all_variables)
        else:
            return traverse(parse_tree.children[0], all_variables) * traverse(parse_tree.children[1], all_variables)

    elif parse_tree.data == 'factor':
        if len(parse_tree.children) == 1:
            return traverse(parse_tree.children[0], all_variables)
        elif parse_tree.children[0] == '-':
            return -traverse(parse_tree.children[1], all_variables)
        elif parse_tree.children[0] == '+':
            return traverse(parse_tree.children[1], all_variables)

    elif parse_tree.data == 'primary':
        if not parse_tree.children[0].__class__ is lark.Token:
            return traverse(parse_tree.children[0], all_variables)
        elif parse_tree.children[0].type == 'RATIONALNUMBER':
            return Solver.get_constant_polynomial(all_variables, str(parse_tree.children[0]))
        else:
            deg = 1
            if len(parse_tree.children) > 1:
                deg = int(parse_tree.children[1])
            degrees = [0] * len(all_variables)
            degrees[find_index_of_variable(
                str(parse_tree.children[0]), all_variables)] = deg
            return Solver.get_degree_polynomial(all_variables, degrees)


def convert_to_desired_poly(poly: Polynomial, program_variables: List[UnknownVariable]) -> Polynomial:
    """
    This function separates program variables from template variables and puts
    template variables in Coefficient.

    Parameters
    ----------
    poly : Polynomial
        Polynomial that should be converted.
    program_variables : List[UnknownVariable]
        List of program variables.

    Returns
    -------
    Polynomial
        Polynomial with separated program variables from template variables.
    """
    monomials = []
    for monomial in poly.monomials:
        coef = monomial.coefficient
        degrees = [0] * len(program_variables)
        for i, var in enumerate(monomial.variables):
            if var in program_variables:
                degrees[find_index_of_variable(
                    var.name, program_variables)] = monomial.degrees[i]
            else:
                for _ in range(monomial.degrees[i]):
                    coef.elements[0].variables.append(var)
        monomials.append(Monomial(program_variables, degrees, coef))
    return Polynomial(program_variables, monomials).revise()


def convert_general_string_to_poly(poly_text: str, all_variables: List[UnknownVariable],
                                   program_variables: List[UnknownVariable]) -> Polynomial:
    """ 
    This function convert a string to Polynomial.

    It uses Lark to parse the string.

    Parameters
    ----------
    poly_text : str
        Text that should be converted to the Polynomial.
    all_variables : List[UnknownVariable]
        List of all variables in the text.
    program_variables : List[UnknownVariable]
        List of program variables.

    Returns
    -------
    Polynomial
        Polynomial that the input string represents.
    """
    parser = Lark(r"""
            start : expression

            expression : term | expression SIGN term 

            term : factor | term "*" factor

            factor : primary | SIGN factor

            primary : VAR | RATIONALNUMBER | VAR "^" RATIONALNUMBER  | "(" expression ")"
            
            SIGN : "+" | "-" 
            VAR: /\w/+
            RATIONALNUMBER : /[+-]?/ NUMBER ("/" NUMBER)?


            %import common.NUMBER
            %import common.NEWLINE -> _NL
            %import common.WS_INLINE
            %ignore WS_INLINE
        """, parser="lalr")

    parse_tree = parser.parse(poly_text)
    return convert_to_desired_poly(traverse(parse_tree, all_variables), program_variables)



import os
import subprocess
from typing import List, Tuple

from .Coefficient import Coefficient, Element, UnknownVariable
from .Constant import Constant, Theorem
from .Constraint import CoefficientConstraint
from .Convertor import Polynomial, convert_general_string_to_poly
from .DNF import DNF
from .Farkas import Farkas
from .Handelman import Handelman
from .Putinar import Putinar
from .Solver import Solver
from .CNF import CNF

class PositiveModel:
    """
    This class is the main class which gets some horn clause as input and find
    the constraints based on the given theorem and find the template value
    using the given solver.

    Parameters
    ----------
    template_variables_name : List[str]
        List of the name of the template variables.
    theorem_name : str
        Name of the algorithm should be used to find the constraints.
    get_SAT : bool, optional
        Should constraint for satisfactory added or not, by default True
    get_UNSAT : bool, optional
        Should constraint for unsatisfactory added or not, by default False
    get_strict : bool, optional
        Should constraint for unsatisfactory in strict form added or not, by default False
    degree_of_sat : int, optional
        Maximum degree of monoids when finding sat constraints in handelman or putinar, by default 0
    degree_of_nonstrict_unsat : int, optional
        Maximum degree of monoids when finding unsat constraints in handelman or putinar, by default 0
    degree_of_strict_unsat : int, optional
        Maximum degree of monoids when finding unsat constraints in strict case in handelman or putinar, by default 0
    max_d_of_strict : int, optional
        Degree of new variable that is generated for strict case in right hand side of equality in putinar, by default 0
    preconditions : List[DNF], optional
        List of conditions that must be satisfied independent of the horn clauses, by default []

    Attributes
    ----------
    paired_constraint : List[Tuple[DNF, DNF, List[UnknownVariable]]]
        List of horn clause constraints.
    template_variables : List[UnknownVariable]
        List of template variables.
    program_variables : List[UnknownVariable]
        List of program variables.
    theorem_name : str
        Name of the algorithm should be used to find the constraints.
    get_SAT : bool
        Should constraint for satisfactory added or not.
    get_UNSAT : bool
        Should constraint for unsatisfactory added or not.
    get_strict : bool
        Should constraint for unsatisfactory in strict form added or not.
    degree_of_sat : int
        Maximum degree of monoids when finding sat constraints in handelman or putinar.
    degree_of_nonstrict_unsat : int
        Maximum degree of monoids when finding unsat constraints in handelman or putinar.
    degree_of_strict_unsat : int
        Maximum degree of monoids when finding unsat constraints in strict case in handelman or putinar.
    max_d_of_strict : int
        Degree of new variable that is generated for strict case in right hand side of equality in putinar.
    preconditions : List[DNF]
        List of conditions that must be satisfied independent of the horn clauses.
    instructions : List[str]
        List of instructions that should be added to the smt file.
    """

    def __init__(self, template_variables_name: List[str],
                 theorem_name: str, get_SAT: bool = True, get_UNSAT: bool = False, get_strict: bool = False,
                 degree_of_sat: int = 0, degree_of_nonstrict_unsat: int = 0, degree_of_strict_unsat: int = 0, max_d_of_strict: int = 0,
                 preconditions: List[DNF] = []
                 ):
        self.paired_constraint = []
        self.template_variables = []
        self.program_variables = []
        for name in template_variables_name:
            self.template_variables.append(UnknownVariable(
                name=name, type_of_var='template_var'))

        self.theorem_name = theorem_name
        self.get_SAT = get_SAT
        self.get_UNSAT = get_UNSAT
        self.get_strict = get_strict
        self.degree_of_sat = degree_of_sat
        self.degree_of_nonstrict_unsat = degree_of_nonstrict_unsat
        self.degree_of_strict_unsat = degree_of_strict_unsat
        self.max_d_of_strict = max_d_of_strict
        self.preconditions = preconditions
        self.instructions = []

    def add_paired_constraint(self, lhs: DNF, rhs: CNF, program_variables: List[UnknownVariable]) -> None:
        """
        Add set of horn clause constraint for lhs => rhs

        Parameters
        ----------
        lhs : DNF
            DNF form of the left hand side of that should be added.
        rhs : DNF
            DNF form of the right hand side of that should be added.
        program_variables : List[UnknownVariable]
            List of program variables.
        """
        for rhsLiteral in rhs.literals:
            new_rhs = CNF([rhsLiteral])
            new_lhs = lhs
            if len(rhsLiteral) > 1:
                for itemInRhsLiteral in rhsLiteral[1:]:
                    new_lhs = new_lhs & DNF([[-itemInRhsLiteral]])
                new_rhs = DNF([[rhsLiteral[0]]])
            for literal in new_lhs.literals:
                for item in new_rhs.literals[0]:
                    self.paired_constraint.append(
                        (literal, item, program_variables))

    def __str__(self) -> str:
        """
        Convert PositiveModel to string.

        Returns
        -------
        str
            String format of the class.
        """
        res = ''
        for pair in self.paired_constraint:
            for lhs_item in pair[0]:
                res += str(lhs_item) + '\n'
            res += '->\n'
            res += str(pair[1]) + '\n'
            res += '----------------------\n'
        return res

    def get_polynomial(self, poly_str: str, program_variables: List[UnknownVariable]) -> Polynomial:
        """
        Generate a polynomial from a given string based on the template and
        program variable in the class.

        Parameters
        ----------
        poly_str : str
            Input string that should be converted to a polynomial.
        program_variables : List[UnknownVariable]
            List of program variables.

        Returns
        -------
        Polynomial
            Polynomial of the given string.
        """
        return convert_general_string_to_poly(poly_str, self.template_variables + program_variables,
                                              program_variables)

    def get_generated_constraints(self) -> List[DNF]:
        """
        This function find the constraint for the list of the class's horn
        clause constraints based on the class configurations.

        Returns
        -------
        List[DNF]
            List of DNF form of constraint for each horn clause.
        """
        all_constraint = []
        for pair in self.paired_constraint:
            theorem = self.theorem_name
            if theorem == 'auto':
                lhs_linear = True
                rhs_linear = pair[1].polynomial.is_linear()
                deg = pair[1].polynomial.get_deg()
                for cons in pair[0]:
                    deg = max(deg, cons.polynomial.get_deg())
                    if not cons.polynomial.is_linear():
                        lhs_linear = False
                if lhs_linear and rhs_linear:
                    theorem = Theorem.FARKAS
                elif lhs_linear:
                    theorem = Theorem.HANDELMAN
                else:
                    theorem = Theorem.PUTINAR
                self.degree_of_sat = deg
                self.degree_of_nonstrict_unsat = deg
                self.degree_of_strict_unsat = deg
                self.max_d_of_strict = deg
            else:
                theorem = Theorem(theorem)

            if theorem == Theorem.FARKAS:
                model = Farkas(variables=pair[2], LHS=pair[0], RHS=pair[1])
            elif theorem == Theorem.HANDELMAN:
                model = Handelman(variables=pair[2], LHS=pair[0], RHS=pair[1],
                                  max_d_for_sat=self.degree_of_sat, max_d_for_unsat=self.degree_of_nonstrict_unsat)
            elif theorem == Theorem.PUTINAR:
                model = Putinar(variables=pair[2], LHS=pair[0], RHS=pair[1],
                                max_d_for_sat=self.degree_of_sat, max_d_for_unsat=self.degree_of_nonstrict_unsat,
                                max_d_for_unsat_strict=self.degree_of_strict_unsat,
                                degree_for_new_var=self.max_d_of_strict)
            else:
                print("no such model")
                return
            new_dnf = []
            if self.get_SAT:
                new_dnf.append(model.get_SAT_constraint())
            if self.get_UNSAT:
                new_dnf.append(model.get_UNSAT_constraint(need_strict=False))
            if self.get_strict:
                if self.theorem_name == 'putinar':
                    for constraint in model.get_UNSAT_constraint(need_strict=True):
                        new_dnf.append(constraint)
                else:
                    new_dnf.append(
                        model.get_UNSAT_constraint(need_strict=True))
            all_constraint.append(DNF(new_dnf))
        return all_constraint

    def create_smt_file(self, output_path: str = "checking.txt", solver_name: str = 'default', solver_path: str = "default",
                        core_iteration_heuristic: bool = False,
                        constant_heuristic: bool = False, real_values: bool = True):
        """
        This function create a smt file based on the given configuration and
        constraints.

        Parameters
        ----------
        output_path : str, optional
            Path to the output file, by default "checking.txt"
        solver_name : str, optional
            Name of the solver, by default 'default'
        solver_path : str, optional
            Path to the solver, by default "default"
        core_iteration_heuristic : bool, optional
            Should core iteration heuristic be applied or not, by default False
        constant_heuristic : bool, optional
            Should removing constant heuristic be applied or not, by default False
        real_values : bool, optional
            Should the variables be integer or real value, by default True
        """
        all_constraint = self.get_generated_constraints()
        if solver_path == "default":
            solver_path = Constant.default_path[solver_name]
        if constant_heuristic and (self.get_SAT ^ self.get_UNSAT ^ self.get_strict) and (
                not (self.get_SAT and self.get_UNSAT and self.get_UNSAT)):
            all_constraint = PositiveModel.remove_equality_constraints(
                all_constraint)
        if core_iteration_heuristic:
            all_constraint = self.core_iteration(all_constraint, solver_path=solver_path,
                                                 real_values=real_values, saving_path=output_path)

        solver_option = Constant.options[solver_name]

        names = ''
        for var in self.template_variables:
            names = names + ' ' + str(var)

        names = names.strip()
        output_command = ''
        if '(check-sat)' in self.instructions:
            output_command += '\n(check-sat)\n'

        if '(get-model)' in self.instructions:
            output_command += f'\n(get-value({names}))\n'

        f = open(output_path, "w")
        f.write(solver_option + Solver.smt_declare_variable_phase(all_constraint, real_values,
                                                                  self.template_variables) + '\n' +
                Solver.convert_constraints_to_smt_format(
                    all_constraint, self.preconditions) + output_command
                )
        f.close()

    def run_on_solver(self, output_path: str = "checking.txt", solver_name: str = 'z3', core_iteration_heuristic: bool = False,
                      constant_heuristic: bool = False, real_values: bool = True) -> Tuple[bool, dict]:
        """ 
        This function finds the constraints for the clauses and runs a solver
        with a given configuration and finds values for the template variables.

        Parameters
        ----------
        output_path : str, optional
            Path to the output file, by default "checking.txt"
        solver_name : str, optional
            Name of the solver, by default 'z3'
        core_iteration_heuristic : bool, optional
            Should core iteration heuristic be applied or not, by default False
        constant_heuristic : bool, optional
            Should removing constant heuristic be applied or not, by default False
        real_values : bool, optional
            Should the variables be integer or real value, by default True

        Returns
        -------
        bool
            True if the model is satisfiable.
        dict
            Dictionary of the template variables and their values.
        """
        solver_path = Constant.default_path[solver_name]
        if solver_path is None:
            print(f"ERROR: Solver {solver_name} is not installed")
            return 'unknown', {}

        self.create_smt_file(output_path, solver_name, solver_path,
                             core_iteration_heuristic, constant_heuristic, real_values)
        output = subprocess.getoutput(
            f'{solver_path} {Constant.command[solver_name]} {output_path}')
        is_sat = output.split('\n')[0]

        values = '\n'.join(output.split('\n')[1:])[1:-1].strip()
        # print(output)
        if is_sat == 'unsupported':
            is_sat = output.split('\n')[1]
            values = '\n'.join(output.split('\n')[2:])[2:-1].strip()
        if is_sat == 'unsat':
            return 'unsat', {}
        if is_sat != 'sat':
            return 'unknown', {}
        values_of_variable = {}

        for line in values.split('\n'):
            line = line.strip()
            line = line[1:-1].strip()
            var_name = line.split(' ')[0]
            var_value = ' '.join(line.split(' ')[1:])
            for temp_var in self.template_variables:
                if temp_var.name == var_name:
                    values_of_variable[temp_var] = var_value
                    break
        result_dictionary = {}
        for var in values_of_variable.keys():
            result_dictionary[var.name] = values_of_variable[var]

        return 'sat', result_dictionary

    @staticmethod
    def get_equality_constraint(all_constraint: List[DNF]) -> CoefficientConstraint:
        """
        Given a list of constraints find a constraint that is in form of a
        variable equal to a constant.

        Parameters
        ----------
        all_constraint : List[DNF]
            List of the constraints.

        Returns
        -------
        CoefficientConstraint
            A equality constraint.
        """
        for dnf in all_constraint:
            for literal in dnf.literals:
                for constraint in literal:
                    if constraint.is_equality():
                        return constraint
        return None

    @staticmethod
    def remove_equality_constraints(all_constraint: List[DNF]) -> List[DNF]:
        """ 
        Remove the equality constraints in each DNF

        Parameters
        ----------
        all_constraint : List[DNF]
            List of the constraints.

        Returns
        -------
        List[DNF]
            New list of constraints after the heuristic is performed.
        """
        while True:
            equality_constraint = PositiveModel.get_equality_constraint(
                all_constraint)
            if equality_constraint is None:
                break
            amount = 0
            if len(equality_constraint.coefficient.elements) == 1:
                variable = equality_constraint.coefficient.elements[0].variables[0]
            else:
                element1 = equality_constraint.coefficient.elements[0]
                element2 = equality_constraint.coefficient.elements[1]
                if len(element1.variables) == 1:
                    variable = element1.variables[0]
                    if element2.constant == 0:
                        amount = 0
                    else:
                        amount = -element2.constant / element1.constant
                else:
                    variable = element2.variables[0]
                    if element1.constant == 0:
                        amount = 0
                    else:
                        amount = -element1.constant / element2.constant

            for dnf in all_constraint:
                for literal in dnf.literals:
                    for constraint in literal:
                        for element in constraint.coefficient.elements:
                            if variable in element.variables:
                                element.variables.remove(variable)
                                element.constant = element.constant * amount
        return all_constraint

    def core_iteration(self, all_constraint: List[DNF], solver_path: str = './solver/z3',
                       saving_path: str = 'save_for_core_iteration_heuristic_temp.txt', real_values: bool = True) -> List[DNF]:
        """
        Perform the core iteration heuristic on set of constraints.

        Parameters
        ----------
        all_constraint : List[DNF]
            List of the constraint that the heuristic should be applied on.
        solver_path : str, optional
            Solver path for finding the core, by default './solver/z3'
        saving_path : str, optional
            Path to a file for temporary saving the output_farkas of solver, by default 'save_for_core_iteration_heuristic_temp.txt'
        real_values : bool, optional
            A boolean that determines if the variables should be integer or real value, by default True

        Returns
        -------
        List[DNF]
            List of constraint after performing the heuristic.
        """
        template_variables = self.template_variables[:]
        unsat = True
        while unsat and len(template_variables) > 0:
            generated_constraint = []
            new_name = []
            for var in template_variables:
                generated_constraint.append(
                    DNF(
                        [[CoefficientConstraint(
                            Coefficient([Element('1', [var])]), '=')]]
                    )
                )

                new_name.append('cons-' + var.name)

            input_of_solver = '(set-option :produce-unsat-cores true)\n'
            input_of_solver += (Solver.smt_declare_variable_phase(
                all_constraint, real_values, self.template_variables))
            input_of_solver += (Solver.convert_constraints_to_smt_format(
                generated_constraint, [], new_name))
            input_of_solver += (Solver.convert_constraints_to_smt_format(
                all_constraint, self.preconditions))
            input_of_solver += '\n(check-sat)\n(get-unsat-core)\n'
            f = open(saving_path, "w")
            f.write(input_of_solver)
            f.close()
            output = subprocess.getoutput(f"{solver_path} {saving_path}")
            sat = output.split()[0]
            core = output.replace('(', ' ').replace(')', ' ').split()[1:]

            os.remove(saving_path)
            if sat == 'sat':
                return generated_constraint + all_constraint

            if len(core) == 0:
                return all_constraint
            for name in core:
                name = name.strip()[5:]
                for var in template_variables:
                    if var.name == name:
                        delete_var = var
                        break
                template_variables.remove(delete_var)
        return all_constraint



import json
import random
import string
import uuid
import os
from typing import Tuple, Union, Callable

from pysmt.fnode import FNode
from pysmt.operators import op_to_str
from pysmt.solvers.solver import Solver as PysmtSolver

from polyqent.Parser import Parser
from polyqent.PositiveModel import PositiveModel


def add_default_config(config: dict) -> dict:
    """
    Add default values to the config dictionary

    Parameters
    ----------
    config : dict
        The config dictionary

    Returns
    -------
    dict
        The config dictionary with default values
    """
    default_config = {
        "SAT_heuristic": True,
        "degree_of_sat": 0,
        "degree_of_nonstrict_unsat": 0,
        "degree_of_strict_unsat": 0,
        "max_d_of_strict": 0,
        "unsat_core_heuristic": False,
        "integer_arithmetic": False,
        "theorem_name": "auto",
        "solver_name": "z3",
        "output_path": "fully_existentially_constraints.txt",
    }

    for key, value in default_config.items():
        config[key] = config.get(key, value)

    return config


def load_config(config_path: str) -> dict:
    """
    Load and parse the config file with default values

    Parameters
    ----------
    config_path : str
        The path to the config file

    Returns
    -------
    dict
        The parsed config file with default values
    """
    try:
        with open(config_path, "r") as file:
            config: dict = json.load(file)
    except:
        config: dict = {}
    return add_default_config(config)


def pysmt_to_smt2(solver: PysmtSolver) -> str:
    """
    Convert a pysmt solver to a smt2 string

    Parameters
    ----------
    solver : pysmt.solvers.solver.Solver
        The pysmt solver

    Returns
    -------
    str
        The smt2 string
    """
    free_vars = set()
    for formula in solver.assertions:
        if formula.is_forall() and not formula.arg(0).is_implies():
            # Check for constraint pairs
            raise ValueError(
                f"PolyQEnt expects universally quantified formulas to consist of constraint pairs in the form `forall x, y, ...: (constraint) => (constraint)`")

        free_vars |= formula.get_free_variables()

    def to_smt2(formula: FNode) -> str:
        
        if formula.is_forall():
            assert len(formula.args()) == 1, f'`Forall` expected 1 argument, got {len(formula.args())}'
            return f'(forall ({" ".join(f"({str(v)} {v.symbol_type()})" for v in formula.quantifier_vars())}) {to_smt2(formula.arg(0))})'
        elif formula.is_implies():
            assert len(formula.args()) == 2, f'`Implies` expected 2 arguments, got {len(formula.args())}'
            return f'(=> {to_smt2(formula.arg(0))} {to_smt2(formula.arg(1))})'
        elif formula.is_and():
            assert len(formula.args()) >= 2, f'`And` expected at least 2 arguments, got {len(formula.args())}'
            return f'(and {" ".join(to_smt2(arg) for arg in formula.args())})'
        elif formula.is_or():
            assert len(formula.args()) >= 2, f'`Or` expected at least 2 arguments, got {len(formula.args())}'
            return f'(or {" ".join(to_smt2(arg) for arg in formula.args())})'
        elif formula.is_le():
            assert len(formula.args()) == 2, f'`Le` expected 2 arguments, got {len(formula.args())}'
            return f'(<= {to_smt2(formula.arg(0))} {to_smt2(formula.arg(1))})'
        elif formula.is_lt():
            assert len(formula.args()) == 2, f'`Lt` expected 2 arguments, got {len(formula.args())}'
            return f'(< {to_smt2(formula.arg(0))} {to_smt2(formula.arg(1))})'
        elif formula.is_equals():
            assert len(formula.args()) == 2, f'`Equals` expected 2 arguments, got {len(formula.args())}'
            return f'(= {to_smt2(formula.arg(0))} {to_smt2(formula.arg(1))})'
        elif formula.is_plus():
            assert len(formula.args()) >= 2, f'`Plus` expected at least 2 arguments, got {len(formula.args())}'
            return f'(+ {" ".join(to_smt2(arg) for arg in formula.args())})'
        elif formula.is_minus():
            assert len(formula.args()) == 2, f'`Minus` expected 2 arguments, got {len(formula.args())}'
            return f'(- {to_smt2(formula.arg(0))} {to_smt2(formula.arg(1))})'
        elif formula.is_times():
            assert len(formula.args()) >= 2, f'`Times` expected at least 2 arguments, got {len(formula.args())}'
            return f'(* {" ".join(to_smt2(arg) for arg in formula.args())})'
        elif formula.is_real_constant():
            return str(formula.constant_value())
        elif formula.is_int_constant():
            return str(formula.constant_value())
        elif formula.is_symbol():
            return str(formula)
        else:
            raise ValueError(f"PolyQEnt does not support '{op_to_str(formula.node_type())}' operator in (sub-)formula {formula}")
    
    constraints = []
    for formula in solver.assertions:
        constraints.append(to_smt2(formula))

    smt2 = '\n'.join(
        [f'(declare-const {var} {var.symbol_type()})' for var in free_vars])
    smt2 += '\n'
    smt2 += '\n'.join([f'(assert {constraint})' for constraint in constraints])
    smt2 += '\n(check-sat)\n(get-model)'
    print(smt2)
    return smt2


def execute(formula: Union[str, PysmtSolver], config: Union[str, dict]) -> Tuple[str, dict]:
    """
    Execute PolyQEnt on the formula with the given configuration
    
    Parameters
    ----------
    formula : Union[str, pysmt.solvers.solver.Solver]
        The formula to execute PolyQEnt on. Either a string to a `.smt2` file
        or a pysmt.Solver object with the constraints already added
    config : Union[str, dict]
        The path to the config file or the parsed config file
        
    Returns
    -------
    str
        The satisfiability of the formula (sat, unsat, unknown)
    dict
        The model of the formula (if it is satisfiable)
    """
    if isinstance(config, str):
        config = load_config(config)
    elif isinstance(config, dict):
        config = add_default_config(config)
    else:
        raise ValueError(
            "Config must be either a path to a config file or a dictionary")

    if isinstance(formula, str):
        with open(formula, "r") as file:
            formula = file.read()
    elif isinstance(formula, PysmtSolver):
        formula = pysmt_to_smt2(formula)
    else:
        raise ValueError(
            "Formula must be either a path to a smt2 file or a pysmt.Solver object")

    return __execute(config, formula, Parser.parse_smt_file)


def execute_smt2(smt2: str, config_path: str) -> Tuple[str, dict]:
    """
    Execute PolyQEnt on the smt2 system

    Parameters
    ----------
    smt2 : str
        The smt2 system
    config_path : str
        The path to the config file

    Returns
    -------
    str
        The satisfiability of the system (sat, unsat, unknown)
    dict
        The model of the system (if it is satisfiable)
    """
    config = load_config(config_path)
    return __execute(config, smt2, Parser.parse_smt_file)


def execute_readable(readable: str, config_path: str) -> Tuple[str, dict]:
    """
    Execute PolyQEnt on the readable system

    Parameters
    ----------
    readable : str
        The readable system
    config_path : str
        The path to the config file

    Returns
    -------
    str
        The satisfiability of the system (sat, unsat, unknown)
    dict
        The model of the system (if it is satisfiable)
    """
    config = load_config(config_path)
    return __execute(config, readable, Parser.parse_readable_file)


def __execute(config: dict, input: str, parser_method: Callable) -> Tuple[str, dict]:
    """
    Execute PolyQEnt on the input system

    Parameters
    ----------
    config : dict
        The dictionary containing configuration information
    input : str
        The input system
    parser_method : Callable
        The method to parse the input

    Returns
    -------
    str
        The satisfiability of the system (sat, unsat, unknown)
    dict
        The model of the system (if it is satisfiable)
    """

    parser = Parser(
        PositiveModel([],
                      config['theorem_name'],
                      True, not config['SAT_heuristic'], not config['SAT_heuristic'],
                      config['degree_of_sat'], config['degree_of_nonstrict_unsat'],
                      config['degree_of_strict_unsat'], config['max_d_of_strict'],
                      preconditions=[],
                      ))

    parser_method(parser, input)
    output_path_exists = True
    try:
        if "output_path" not in config:
            output_path_exists = False
            config["output_path"] = './POLYQENT_delme_' + str(uuid.uuid4()) + ''.join(
                random.choices(string.ascii_uppercase + string.digits, k=9))
            with open(config["output_path"], 'x') as file:
                file.write("")
        print("Running solver...")
        sat, model = parser.model.run_on_solver(output_path=config["output_path"], solver_name=config["solver_name"],
                                                core_iteration_heuristic=config['unsat_core_heuristic'],
                                                constant_heuristic=False,
                                                real_values=not config['integer_arithmetic'])
    finally:
        if not output_path_exists:
            os.remove(config["output_path"])
    return sat, model


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--smt2", type=str, help="Path to the smt2 file")
    parser.add_argument("--readable", type=str,
                        help="Path to the readable file")
    parser.add_argument("--config", type=str,
                        help="Path to the config file")
    args = parser.parse_args()

    if args.smt2:
        with open(args.smt2, "r") as file:
            smt2 = file.read()
        is_sat, model = execute_smt2(smt2, args.config)
    elif args.readable:
        with open(args.readable, "r") as file:
            readable = file.read()
        is_sat, model = execute_readable(readable, args.config)
    else:
        raise ValueError("Either --smt2 or --readable must be provided")

    print(f"The system is {is_sat}")
    if is_sat == 'sat':
        print("Model:")
        for var, value in model.items():
            print(f"{var}: {value}")



from typing import List, Tuple

from .Constraint import CoefficientConstraint, PolynomialConstraint
from .Polynomial import Polynomial
from .Solver import Solver
from .UnknownVariable import UnknownVariable


class Farkas:
    """
    A class that performs the constraint transformation according to Farkas
    Lemma.

    Parameters
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm

    Attributes
    ----------
    variables : List[UnknownVariable]
        The list of variables of the polynomials.
    LHS : List[PolynomialConstraint]
        The list of left hand side constraints in farkas algorithm
    RHS : PolynomialConstraint
        The right hand side constraint in algorithm
    """

    def __init__(self, variables: List[UnknownVariable], LHS: List[PolynomialConstraint], RHS: PolynomialConstraint):
        self.variables = variables
        self.RHS = RHS
        self.LHS = LHS

    def get_poly_sum(self, need_strict: bool = False) -> Tuple[Polynomial, CoefficientConstraint]:
        """ 
        This function returns a polynomial y_0 + y_1*f_1 + y_2*f_2 ...+ y_n*f_n
        where f_i are left hand side polynomials and y_i are newly created 
        variable in farkas theorem.

        Parameters
        ----------
        need_strict : bool, optional
            Is it generated for the 0 > 0 case or not, by default False

        Returns
        -------
        Tuple[Polynomial, CoefficientConstraint]
            polynomial of sum of all left hand side with new variables and corresponding constraints.
        """
        new_var = Solver.get_variable_polynomial(
            self.variables, 'y0', 'generated_for_Farkas')
        polynomial_of_sum = new_var

        sum_of_strict = Solver.get_constant_polynomial(
            self.variables, '0') + new_var

        constraints = [CoefficientConstraint(
            new_var.monomials[0].coefficient, '>=')]

        for i, left_constraint in enumerate(self.LHS):
            left_poly = left_constraint.polynomial
            new_var_poly = Solver.get_variable_polynomial(self.variables, f'y{i + 1}',
                                                          'generated_for_Farkas')
            if left_constraint.is_strict():
                sum_of_strict = sum_of_strict + new_var_poly

            polynomial_of_sum = polynomial_of_sum + (new_var_poly * left_poly)
            constraints.append(CoefficientConstraint(
                new_var_poly.monomials[0].coefficient, '>='))

        if need_strict or self.RHS.is_strict():
            constraints.append(CoefficientConstraint(
                sum_of_strict.monomials[0].coefficient, '>'))
        return polynomial_of_sum, constraints

    def get_SAT_constraint(self) -> List[CoefficientConstraint]:
        """ 
        A function to find the constraints when the LHS => RHS is satisfiable.

        Returns
        -------
        List[CoefficientConstraint]
            list of coefficient constraints when it is satisfiable
        """
        polynomial_of_sum, constraints = self.get_poly_sum()
        return Solver.find_equality_constraint(polynomial_of_sum, self.RHS.polynomial) + constraints

    def get_UNSAT_constraint(self, need_strict: bool = False) -> List[CoefficientConstraint]:
        """ 
        A function to find the constraints when it is not satisfiable.

        Two set of constraint can be generated:
            1. `LHS => -1 >= 0`
            2. `LHS => 0 > 0`

        Parameters
        ----------
        need_strict : bool, optional
            determine which set of constraint to generate, by default False

        Returns
        -------
        List[CoefficientConstraint]
            list of coefficient constraints when it is not satisfiable
        """
        if need_strict:
            polynomial_of_sum, constraints = self.get_poly_sum(need_strict)
            return Solver.find_equality_constraint(polynomial_of_sum,
                                                   Polynomial(polynomial_of_sum.variables, [])) + constraints
        polynomial_of_sum, constraints = self.get_poly_sum()
        return Solver.find_equality_constraint(polynomial_of_sum,
                                               Solver.get_constant_polynomial(self.RHS.polynomial.variables, '-1')) + constraints



from typing import Any

import lark
from lark import Lark

from .Constraint import PolynomialConstraint
from .Convertor import convert_to_desired_poly, find_index_of_variable
from .DNF import DNF
from .PositiveModel import PositiveModel
from .Solver import Solver
from .UnknownVariable import UnknownVariable
from .CNF import CNF


class Parser:
    """
    This class is used to parse the input file and create a PositiveModel object.

    Parameters
    ----------
    model : PositiveModel
        The model that will be created.

    Attributes
    ----------
    model : PositiveModel
        The model that will be created.
    """

    def __init__(self, model: PositiveModel):
        self.model = model

    def traverse_readable_tree(self, parse_tree: lark.tree.Tree) -> Any:
        """
        This function traverses the parse tree generated by lark and returns
        the corresponding object.

        Parameters
        ----------
        parse_tree : lark.tree.Tree
            A node in parse tree.(starting from root)

        Returns
        -------
        Any
            The object that the parse tree represents.
        """
        if parse_tree.data == 'start':
            for child in parse_tree.children:
                self.traverse_readable_tree(child)
        elif parse_tree.data == 'program_var':
            for child in parse_tree.children:
                self.model.program_variables.append(
                    UnknownVariable(str(child), type_of_var='program_var'))

        elif parse_tree.data == 'template_var':
            for child in parse_tree.children:
                self.model.template_variables.append(
                    UnknownVariable(str(child), type_of_var='template_var'))

        elif parse_tree.data == 'hornclause':
            lhs = self.traverse_readable_tree(parse_tree.children[0])
            rhs = self.traverse_readable_tree(parse_tree.children[1])
            self.model.add_paired_constraint(lhs, rhs)
            return
        elif parse_tree.data == 'precondition':
            dnf = self.traverse_readable_tree(parse_tree.children[0])
            self.model.preconditions.append(dnf)
            return
        elif parse_tree.data == 'dnf':
            if len(parse_tree.children) == 1 and parse_tree.children[0].data == 'constraint':
                return DNF([[self.traverse_readable_tree(parse_tree.children[0])]])
            elif len(parse_tree.children) == 1:
                return self.traverse_readable_tree(parse_tree.children[0])
            else:
                if str(parse_tree.children[1]) == "AND":
                    return (self.traverse_readable_tree(parse_tree.children[0]) & self.traverse_readable_tree(
                        parse_tree.children[2]))
                else:
                    return (self.traverse_readable_tree(parse_tree.children[0]) | self.traverse_readable_tree(
                        parse_tree.children[2]))
        elif parse_tree.data == 'cnf':
            if len(parse_tree.children) == 1 and parse_tree.children[0].data == 'constraint':
                return CNF([[self.traverse_readable_tree(parse_tree.children[0])]])
            elif len(parse_tree.children) == 1:
                return self.traverse_readable_tree(parse_tree.children[0])
            else:
                if str(parse_tree.children[1]) == "AND":
                    return (self.traverse_readable_tree(parse_tree.children[0]) & self.traverse_readable_tree(
                        parse_tree.children[2]))
                else:
                    return (self.traverse_readable_tree(parse_tree.children[0]) | self.traverse_readable_tree(
                        parse_tree.children[2]))

        elif parse_tree.data == 'literal':
            literal = []
            for i in range(len(parse_tree.children)):
                literal.append(self.traverse_readable_tree(
                    parse_tree.children[i]))
            return literal
        elif parse_tree.data == 'constraint':
            return PolynomialConstraint(
                self.traverse_readable_tree(parse_tree.children[0]) - self.traverse_readable_tree(
                    parse_tree.children[2]),
                str(parse_tree.children[1]))
        elif parse_tree.data == 'polynomial':
            return convert_to_desired_poly(self.traverse_readable_tree(parse_tree.children[0]),
                                           self.model.program_variables)
        elif parse_tree.data == 'expression':
            if len(parse_tree.children) == 1:
                return self.traverse_readable_tree(parse_tree.children[0])
            elif parse_tree.children[1] == '+':
                return self.traverse_readable_tree(parse_tree.children[0]) + self.traverse_readable_tree(
                    parse_tree.children[2])
            elif parse_tree.children[1] == '-':
                return self.traverse_readable_tree(parse_tree.children[0]) - self.traverse_readable_tree(
                    parse_tree.children[2])

        elif parse_tree.data == 'term':
            if len(parse_tree.children) == 1:
                return self.traverse_readable_tree(parse_tree.children[0])
            else:
                return self.traverse_readable_tree(parse_tree.children[0]) * self.traverse_readable_tree(
                    parse_tree.children[1])

        elif parse_tree.data == 'factor':
            if len(parse_tree.children) == 1:
                return self.traverse_readable_tree(parse_tree.children[0])
            elif parse_tree.children[0] == '-':
                return -self.traverse_readable_tree(parse_tree.children[1])
            elif parse_tree.children[0] == '+':
                return self.traverse_readable_tree(parse_tree.children[1])

        elif parse_tree.data == 'primary':
            if not parse_tree.children[0].__class__ is lark.Token:
                return self.traverse_readable_tree(parse_tree.children[0])
            elif parse_tree.children[0].type == 'RATIONALNUMBER':
                return Solver.get_constant_polynomial(self.model.template_variables + self.model.program_variables,
                                                      str(parse_tree.children[0]))
            else:
                deg = 1
                if len(parse_tree.children) > 1:
                    deg = int(parse_tree.children[1])
                degrees = [
                    0] * len(self.model.template_variables + self.model.program_variables)
                degrees[find_index_of_variable(str(parse_tree.children[0]),
                                               self.model.template_variables + self.model.program_variables)] = deg
                return Solver.get_degree_polynomial(self.model.template_variables + self.model.program_variables,
                                                    degrees)

    def parse_readable_file(self, poly_text: str) -> None:
        """
        This function parses the input file and creates a PositiveModel object.

        Parameters
        ----------
        poly_text : str
            The content of the input file.
        """
        parser = Lark(r"""
                start : program_var template_var precondition* hornclause*

                program_var : "Program_var:" VAR* ";"

                template_var : "Template_var:" VAR* ";"

                precondition : "Precondition:" dnf

                hornclause : "Horn_clause:" dnf "->" cnf

                dnf : constraint | "(" dnf ")" | dnf LOGICAL_SIGN dnf
                cnf : constraint | "(" cnf ")" | cnf LOGICAL_SIGN cnf

                constraint : polynomial COMP_SIGN polynomial 
                polynomial : expression
                expression : term | expression SIGN term 

                term : factor | term "*" factor

                factor : primary | SIGN factor

                primary : VAR | RATIONALNUMBER | VAR "^" RATIONALNUMBER  | "(" expression ")"

                LOGICAL_SIGN : "AND" | "OR" 
                COMP_SIGN : ">" | "=" | "<" | ">=" | "<="
                SIGN : "+" | "-" 
                VAR: /[a-zA-Z0-9_]/+
                RATIONALNUMBER : /[+-]?/ NUMBER ("/" NUMBER)?


                %import common.NUMBER
                %import common.NEWLINE -> _NL
                %import common.WS_INLINE
                %import common.WS
                %ignore WS
            """, parser="lalr")

        parse_tree = parser.parse(poly_text)
        self.traverse_readable_tree(parse_tree)

    def traverse_smt_tree(self, parse_tree: lark.tree.Tree) -> Any:
        """
        This function traverses the parse tree generated by lark and returns
        the corresponding object.

        Parameters
        ----------
        parse_tree : lark.tree.Tree
            A node in parse tree.(starting from root)

        Returns
        -------
        Any
            The object that the parse tree represents.
        """
        if parse_tree.data == 'start':
            for child in parse_tree.children:
                self.traverse_smt_tree(child)

        elif parse_tree.data == 'instructions':
            self.model.instructions.append(parse_tree.children[0])
            return
        elif parse_tree.data == 'declare_var':
            self.model.template_variables.append(
                UnknownVariable(str(parse_tree.children[0]), type_of_var='template_var'))

        elif parse_tree.data == 'assertion':
            self.traverse_smt_tree(parse_tree.children[0])
        elif parse_tree.data == 'hornclause':
            self.model.program_variables = []
            for i in range(len(parse_tree.children) - 2):
                self.traverse_smt_tree(parse_tree.children[i])
            lhs = self.traverse_smt_tree(parse_tree.children[-2])
            rhs = self.traverse_smt_tree(parse_tree.children[-1])
            self.model.add_paired_constraint(
                lhs, rhs, self.model.program_variables)
            return
        elif parse_tree.data == 'program_variables':
            self.model.program_variables.append(UnknownVariable(
                str(parse_tree.children[0]), type_of_var='program_var'))
        elif parse_tree.data == 'precondition':
            dnf = self.traverse_smt_tree(parse_tree.children[0])
            if len(parse_tree.children) == 2:
                second_dnf = self.traverse_smt_tree(parse_tree.children[1])
                self.model.preconditions.append((dnf, second_dnf))
            else:
                self.model.preconditions.append((dnf,))
            return
        elif parse_tree.data == 'dnf':
            if len(parse_tree.children) == 1:
                return self.traverse_smt_tree(parse_tree.children[0])
            else:
                if str(parse_tree.children[0]) == "and":
                    result_dnf = DNF([])
                    for i in range(1, len(parse_tree.children)):
                        result_dnf = result_dnf & self.traverse_smt_tree(
                            parse_tree.children[i])
                    return result_dnf
                else:
                    result_dnf = DNF([])
                    for i in range(1, len(parse_tree.children)):
                        result_dnf = result_dnf | self.traverse_smt_tree(
                            parse_tree.children[i])
                    return result_dnf
        elif parse_tree.data == 'cnf':
            if len(parse_tree.children) == 1:
                result = CNF([])
                for literal in self.traverse_smt_tree(parse_tree.children[0]).literals[0]:
                    result = result & CNF([[literal]])
                return result
            else:
                if str(parse_tree.children[0]) == "and":
                    result_cnf = CNF([])
                    for i in range(1, len(parse_tree.children)):
                        result_cnf = result_cnf & self.traverse_smt_tree(
                            parse_tree.children[i])
                    return result_cnf
                else:
                    result_cnf = CNF([])
                    for i in range(1, len(parse_tree.children)):
                        result_cnf = result_cnf | self.traverse_smt_tree(
                            parse_tree.children[i])
                    return result_cnf

        elif parse_tree.data == 'constraint':
            if parse_tree.children[0] == '=':
                return DNF(
                    [[
                        PolynomialConstraint(
                            self.traverse_smt_tree(parse_tree.children[1])
                            -
                            self.traverse_smt_tree(parse_tree.children[2]),
                            '>='),
                        PolynomialConstraint(
                            self.traverse_smt_tree(parse_tree.children[1])
                            -
                            self.traverse_smt_tree(parse_tree.children[2]),
                            '<='),

                    ]])
            return DNF(
                [[
                    PolynomialConstraint(
                        self.traverse_smt_tree(parse_tree.children[1])
                        -
                        self.traverse_smt_tree(parse_tree.children[2]),
                        str(parse_tree.children[0]))]])
        elif parse_tree.data == 'polynomial':
            return convert_to_desired_poly(self.traverse_smt_tree(parse_tree.children[0]), self.model.program_variables)
        elif parse_tree.data == 'expression':

            if len(parse_tree.children) == 1:
                if parse_tree.children[0].data == "fraction":
                    return Solver.get_constant_polynomial(self.model.template_variables + self.model.program_variables,
                                                          self.traverse_smt_tree(
                                                              parse_tree.children[0])
                                                          )
                return self.traverse_smt_tree(parse_tree.children[0])
            elif len(parse_tree.children) == 2:
                if str(parse_tree.children[0]) == '+':
                    return self.traverse_smt_tree(parse_tree.children[1])
                elif str(parse_tree.children[0]) == '-':
                    return -self.traverse_smt_tree(parse_tree.children[1])
            elif len(parse_tree.children) == 3:

                if str(parse_tree.children[0]) == '+':
                    return self.traverse_smt_tree(parse_tree.children[1]) + self.traverse_smt_tree(
                        parse_tree.children[2])
                elif str(parse_tree.children[0]) == '-':
                    return self.traverse_smt_tree(parse_tree.children[1]) - self.traverse_smt_tree(
                        parse_tree.children[2])
                elif str(parse_tree.children[0]) == '*':
                    return self.traverse_smt_tree(parse_tree.children[1]) * self.traverse_smt_tree(
                        parse_tree.children[2])
            else:
                poly = self.traverse_smt_tree(parse_tree.children[1])
                for i in range(2, len(parse_tree.children)):
                    if str(parse_tree.children[0]) == '+':
                        poly = poly + \
                            self.traverse_smt_tree(parse_tree.children[i])
                    else:
                        poly = poly * \
                            self.traverse_smt_tree(parse_tree.children[i])
                return poly
        elif parse_tree.data == 'primary':
            if type(parse_tree.children[0]) is lark.tree.Tree:
                return Solver.get_constant_polynomial(self.model.template_variables + self.model.program_variables,
                                                      self.traverse_smt_tree(parse_tree.children[0]))

            if parse_tree.children[0].type == 'VAR':
                deg = 1
                if len(parse_tree.children) > 1:
                    deg = int(parse_tree.children[1])
                degrees = [
                    0] * len(self.model.template_variables + self.model.program_variables)
                degrees[find_index_of_variable(str(parse_tree.children[0]),
                                               self.model.template_variables + self.model.program_variables)] = deg
                return Solver.get_degree_polynomial(self.model.template_variables + self.model.program_variables,
                                                    degrees)

        elif parse_tree.data == 'fraction':
            return str(self.traverse_smt_tree(parse_tree.children[0])) + '/' + str(
                self.traverse_smt_tree(parse_tree.children[1]))
        elif parse_tree.data == 'rationalnumber':
            if len(parse_tree.children) == 1:
                return str(parse_tree.children[0])
            if len(parse_tree.children) == 2:
                return str(parse_tree.children[0]) + str(parse_tree.children[1])

    def parse_smt_file(self, poly_text: str) -> None:
        """
        This function parses the input file and creates a PositiveModel object.

        Parameters
        ----------
        poly_text : str
            The content of the input file.
        """
        parser = Lark(r"""
                start : declare_var* assertion* instructions* 

                instructions: INS
                INS : "(check-sat)" | "(get-model)"
                declare_var: "(declare-const" VAR VAR_TYPE ")"

                assertion: "(assert" precondition  ")" | "(assert" hornclause  ")"
                precondition : dnf | "(=>" dnf dnf ")" 

                hornclause : "(forall" "(" program_variables* ")" "(=>" dnf cnf ")" ")"
                program_variables : "(" VAR VAR_TYPE ")" 
                dnf : constraint | "(" LOGICAL_SIGN dnf+ ")" 
                cnf : constraint | "(" LOGICAL_SIGN cnf+ ")" 

                constraint : "(" COMP_SIGN polynomial polynomial ")" 
                polynomial : expression
                expression : "(/" fraction ")" | primary | "(" SIGN expression ")" | "(" SIGN expression+ ")"

                primary : VAR | rationalnumber 

                LOGICAL_SIGN : "and" | "or" 
                COMP_SIGN : ">" | "=" | "<" | ">=" | "<="
                SIGN : "+" | "-" | "*"
                VAR: /[a-zA-Z0-9_]/+
                VAR_TYPE: "Int" | "Real" 
                rationalnumber : NUMBER | "(" SIGN NUMBER ")" | SIGN NUMBER 
                fraction: rationalnumber  rationalnumber 

                %import common.NUMBER
                %import common.NEWLINE -> _NL
                %import common.WS_INLINE
                %import common.WS
                %ignore WS
            """, parser="lalr")

        parse_tree = parser.parse(poly_text)
        self.traverse_smt_tree(parse_tree)



from __future__ import annotations

from typing import List


class DNF:
    """
    A class that represents disjunctive normal form.

    Parameters
    ----------
    literals : List
        List of literals that are in DNF.

    Attributes
    ----------
    literals : List
        List of literals that are in DNF.
    """

    def __init__(self, literals: List):
        self.literals = literals

    def __or__(self, other: DNF) -> DNF:
        """ 
        Or two DNF.

        Or of two DNF is the union of literals of each.

        Parameters
        ----------
        other : DNF
            The other DNF that should be or with this class.

        Returns
        -------
        DNF
            New DNF that is or of two DNF.
        """
        return DNF(self.literals + other.literals)

    def __and__(self, other: DNF) -> DNF:
        """ 
        And two DNF.

        The result is a DNF where each of the literal is the union of two literal in each DNF.

        Parameters
        ----------
        other : DNF
            The other DNF that should be and with this class.

        Returns
        -------
        DNF
            New DNF that is and of two DNF.
        """
        if len(self.literals) == 0:
            return DNF(other.literals)
        if len(other.literals) == 0:
            return DNF(self.literals)

        literal_list = []
        for first_literal in self.literals:
            for second_literal in other.literals:
                literal_list.append(first_literal + second_literal)
        return DNF(literal_list)

    def __neg__(self) -> DNF:
        """ 
        Negate a DNF

        For negating a DNF it is sufficient to negate all its literal and And 
        them together. Negate a literal makes a DNF.

        Returns
        -------
        DNF
            New DNF that is negation of the previous DNF.
        """
        result_DNF = DNF([])
        for literal in self.literals:
            new_arr = []
            for item in literal:
                new_arr.append([-item])
            result_DNF = result_DNF & DNF(new_arr)
        return result_DNF

    def __str__(self) -> str:
        """
        Convert DNF to string.

        Returns
        -------
        str
            String format of the DNF.
        """
        res = ''
        for literal in self.literals:
            res += '\n  AND '.join(["\t" + str(item) for item in literal])
            res += '\n OR \n'
        return '(\n' + res + ')\n'

    def convert_to_preorder(self) -> str:
        """
        Convert DNF to preorder format.

        Returns
        -------
        str
            String in preorder format of the DNF.
        """
        res = '( or '
        for literal in self.literals:
            res += '( and '
            for item in literal:
                res += item.convert_to_preorder()
                res += ' '
            res += ') '
        res += ' )'
        return res
